import os, sys, pprint, re, json, pathlib, hashlib, subprocess, glob, tempfile

executePath = os.getcwd()
sys.dont_write_bytecode = True
scriptPath = os.path.dirname(os.path.realpath(__file__))
sys.path.append(scriptPath + '/..')
import qt_version

def finish(code):
    global executePath
    os.chdir(executePath)
    sys.exit(code)

def error(text):
    print('[ERROR] ' + text)
    finish(1)

def nativeToolsError():
    error('Make sure to run from Native Tools Command Prompt.')

win = (sys.platform == 'win32')
mac = (sys.platform == 'darwin')

if win and not 'Platform' in os.environ:
    nativeToolsError()

win32 = win and (os.environ['Platform'] == 'x86')
win64 = win and (os.environ['Platform'] == 'x64')
winarm = win and (os.environ['Platform'] == 'arm64')

arch = ''
if win32:
    arch = 'x86'
elif win64:
    arch = 'x64'
elif winarm:
    arch = 'arm'
if not qt_version.resolve(arch):
    error('Usupported platform.')

qt = os.environ.get('QT')

if win and not 'COMSPEC' in os.environ:
    error('COMSPEC environment variable is not set.')

if win and not win32 and not win64 and not winarm:
    nativeToolsError()

os.chdir(scriptPath + '/../../../..')

pathSep = ';' if win else ':'
libsLoc = 'Libraries' if not win64 else (os.path.join('Libraries', 'win64'))
keysLoc = 'cache_keys'

rootDir = os.getcwd()
libsDir = os.path.realpath(os.path.join(rootDir, libsLoc))
thirdPartyDir = os.path.realpath(os.path.join(rootDir, 'ThirdParty'))
usedPrefix = os.path.realpath(os.path.join(libsDir, 'local'))

optionsList = [
    'qt6',
    'skip-release',
    'build-stackwalk',
]
options = []
runCommand = []
customRunCommand = False
for arg in sys.argv[1:]:
    if customRunCommand:
        runCommand.append(arg)
    if arg in optionsList:
        options.append(arg)
    elif arg == 'run':
        customRunCommand = True
    elif arg == 'shell':
        customRunCommand = True
        runCommand.append('shell')

if not os.path.isdir(os.path.join(libsDir, keysLoc)):
    pathlib.Path(os.path.join(libsDir, keysLoc)).mkdir(parents=True, exist_ok=True)
if not os.path.isdir(os.path.join(thirdPartyDir, keysLoc)):
    pathlib.Path(os.path.join(thirdPartyDir, keysLoc)).mkdir(parents=True, exist_ok=True)

pathPrefixes = [
    'ThirdParty\\msys64\\mingw64\\bin',
    'ThirdParty\\jom',
    'ThirdParty\\cmake\\bin',
    'ThirdParty\\gyp',
] if win else [
    'ThirdParty/gyp',
    'ThirdParty/yasm',
    'ThirdParty/depot_tools',
]
pathPrefix = ''
for singlePrefix in pathPrefixes:
    pathPrefix = pathPrefix + os.path.join(rootDir, singlePrefix) + pathSep

environment = {
    'USED_PREFIX': usedPrefix,
    'ROOT_DIR': rootDir,
    'LIBS_DIR': libsDir,
    'THIRDPARTY_DIR': thirdPartyDir,
    'PATH_PREFIX': pathPrefix,
}
if (win32):
    environment.update({
        'SPECIAL_TARGET': 'win',
        'X8664': 'x86',
        'WIN32X64': 'Win32',
    })
elif (win64):
    environment.update({
        'SPECIAL_TARGET': 'win64',
        'X8664': 'x64',
        'WIN32X64': 'x64',
    })
elif (winarm):
    environment.update({
        'SPECIAL_TARGET': 'winarm',
        'X8664': 'ARM64',
        'WIN32X64': 'ARM64',
    })
elif (mac):
    environment.update({
        'SPECIAL_TARGET': 'mac',
        'MAKE_THREADS_CNT': '-j' + str(os.cpu_count()),
        'MACOSX_DEPLOYMENT_TARGET': '10.13',
        'UNGUARDED': '-Werror=unguarded-availability-new',
        'MIN_VER': '-mmacosx-version-min=10.13',
    })

ignoreInCacheForThirdParty = [
    'USED_PREFIX',
    'LIBS_DIR',
    'SPECIAL_TARGET',
    'X8664',
    'WIN32X64',
]

environmentKeyString = ''
envForThirdPartyKeyString = ''
for key in environment:
    part = key + '=' + environment[key] + ';'
    environmentKeyString += part
    if not key in ignoreInCacheForThirdParty:
        envForThirdPartyKeyString += part
environmentKey = hashlib.sha1(environmentKeyString.encode('utf-8')).hexdigest()
envForThirdPartyKey = hashlib.sha1(envForThirdPartyKeyString.encode('utf-8')).hexdigest()

modifiedEnv = os.environ.copy()
for key in environment:
    modifiedEnv[key] = environment[key]

modifiedEnv['PATH'] = environment['PATH_PREFIX'] + modifiedEnv['PATH']

def computeFileHash(path):
    sha1 = hashlib.sha1()
    with open(path, 'rb') as f:
        while True:
            data = f.read(256 * 1024)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()

def computeCacheKey(stage):
    if (stage['location'] == 'ThirdParty'):
        envKey = envForThirdPartyKey
    else:
        envKey = environmentKey
    objects = [
        envKey,
        stage['location'],
        stage['name'],
        stage['version'],
        stage['commands']
    ]
    for pattern in stage['dependencies']:
        pathlist = glob.glob(os.path.join(libsDir, pattern))
        items = [pattern]
        if len(pathlist) == 0:
            pathlist = glob.glob(os.path.join(thirdPartyDir, pattern))
        if len(pathlist) == 0:
            error('Nothing found: ' + pattern)
        for path in pathlist:
            if not os.path.exists(path):
                error('Not found: ' + path)
            items.append(computeFileHash(path))
        objects.append(':'.join(items))
    return hashlib.sha1(';'.join(objects).encode('utf-8')).hexdigest()

def keyPath(stage):
    return os.path.join(stage['directory'], keysLoc, stage['name'])

def checkCacheKey(stage):
    if not 'key' in stage:
        error('Key not set in stage: ' + stage['name'])
    key = keyPath(stage)
    if not os.path.exists(os.path.join(stage['directory'], stage['name'])):
        return 'NotFound'
    if not os.path.exists(key):
        return 'Stale'
    with open(key, 'r') as file:
        return 'Good' if (file.read() == stage['key']) else 'Stale'

def clearCacheKey(stage):
    key = keyPath(stage)
    if os.path.exists(key):
        os.remove(key)

def writeCacheKey(stage):
    if not 'key' in stage:
        error('Key not set in stage: ' + stage['name'])
    key = keyPath(stage)
    with open(key, 'w') as file:
        file.write(stage['key'])

stages = []

def removeDir(folder):
    if win:
        return 'if exist ' + folder + ' rmdir /Q /S ' + folder + '\nif exist ' + folder + ' exit /b 1'
    return 'rm -rf ' + folder

def setVar(key, multilineValue):
    singlelineValue = ' '.join(multilineValue.replace('\n', '').split());
    if win:
        return 'SET "' + key + '=' + singlelineValue + '"';
    return key + '="' + singlelineValue + '"';

def filterByPlatform(commands):
    commands = commands.split('\n')
    result = ''
    dependencies = []
    version = '0'
    skip = False
    for command in commands:
        m = re.match(r'(!?)([a-z0-9_]+):', command)
        if m and m.group(2) != 'depends' and m.group(2) != 'version':
            scopes = m.group(2).split('_')
            inscope = 'common' in scopes
            if win and 'win' in scopes:
                inscope = True
            if win32 and 'win32' in scopes:
                inscope = True
            if win64 and 'win64' in scopes:
                inscope = True
            if winarm and 'winarm' in scopes:
                inscope = True
            if mac and 'mac' in scopes:
                inscope = True
            # if linux and 'linux' in scopes:
            #     inscope = True
            if 'release' in scopes:
                if 'skip-release' in options:
                    inscope = False
                elif len(scopes) == 1:
                    continue
            skip = inscope if m.group(1) == '!' else not inscope
        elif not skip and not re.match(r'\s*#', command):
            if m and m.group(2) == 'version':
                version = version + '.' + command[len(m.group(0)):].strip()
            elif m and m.group(2) == 'depends':
                pattern = command[len(m.group(0)):].strip()
                dependencies.append(pattern)
            else:
                command = command.strip()
                if len(command) > 0:
                    result = result + command + '\n'
    return [result, dependencies, version]

def stage(name, commands, location = 'Libraries'):
    if location == 'Libraries':
        directory = libsDir
    elif location == 'ThirdParty':
        directory = thirdPartyDir
    else:
        error('Unknown location: ' + location)
    [commands, dependencies, version] = filterByPlatform(commands)
    if len(commands) > 0:
        stages.append({
            'name': name,
            'location': location,
            'directory': directory,
            'commands': commands,
            'version': version,
            'dependencies': dependencies
        })

def winFailOnEach(command):
    commands = command.split('\n')
    result = ''
    startingCommand = True
    for command in commands:
        command = re.sub(r'\$([A-Za-z0-9_]+)', r'%\1%', command)
        if re.search(r'\$[^<]', command):
            error('Bad command: ' + command)
        appendCall = startingCommand and not re.match(r'(if|for) ', command)
        called = 'call ' + command if appendCall else command
        result = result + called
        if command.endswith('^'):
            startingCommand = False
        else:
            startingCommand = True
            result = result + '\r\nif %errorlevel% neq 0 exit /b %errorlevel%\r\n'
    return result

def printCommands(commands):
    print('---------------------------------COMMANDS-LIST----------------------------------')
    print(commands, end='')
    print('--------------------------------------------------------------------------------')

def run(commands):
    printCommands(commands)
    if win:
        if os.path.exists("command.bat"):
            os.remove("command.bat")
        with open("command.bat", 'w') as file:
            file.write('@echo OFF\r\n' + winFailOnEach(commands))
        result = subprocess.run("command.bat", shell=True, env=modifiedEnv).returncode == 0
        if result and os.path.exists("command.bat"):
            os.remove("command.bat")
        return result
    elif re.search(r'\%', commands):
        error('Bad command: ' + commands)
    else:
        return subprocess.run("set -e\n" + commands, shell=True, env=modifiedEnv).returncode == 0

# Thanks https://stackoverflow.com/a/510364
class _Getch:
    """Gets a single character from standard input.  Does not echo to the
screen."""
    def __init__(self):
        try:
            self.impl = _GetchWindows()
        except ImportError:
            self.impl = _GetchUnix()

    def __call__(self): return self.impl()

class _GetchUnix:
    def __init__(self):
        import tty, sys

    def __call__(self):
        import sys, tty, termios
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch

class _GetchWindows:
    def __init__(self):
        import msvcrt

    def __call__(self):
        import msvcrt
        return msvcrt.getch().decode('ascii')

getch = _Getch()

def runStages():
    onlyStages = []
    rebuildStale = False
    for arg in sys.argv[1:]:
        if arg in options:
            continue
        elif arg == 'silent':
            rebuildStale = True
            continue
        found = False
        for stage in stages:
            if stage['name'] == arg:
                onlyStages.append(arg)
                found = True
                break
        if not found:
            error('Unknown argument: ' + arg)
    count = len(stages)
    index = 0
    for stage in stages:
        if len(onlyStages) > 0 and not stage['name'] in onlyStages:
            continue
        index = index + 1
        version = ('#' + str(stage['version'])) if (stage['version'] != '0') else ''
        prefix = '[' + str(index) + '/' + str(count) + '](' + stage['location'] + '/' + stage['name'] + version + ')'
        print(prefix + ': ', end = '', flush=True)
        stage['key'] = computeCacheKey(stage)
        commands = removeDir(stage['name']) + '\n' + stage['commands']
        checkResult = 'Forced' if len(onlyStages) > 0 else checkCacheKey(stage)
        if checkResult == 'Good':
            print('SKIPPING')
            continue
        elif checkResult == 'NotFound':
            print('NOT FOUND, ', end='')
        elif checkResult == 'Stale' or checkResult == 'Forced':
            if checkResult == 'Stale':
                print('CHANGED, ', end='')
            if rebuildStale:
                checkResult == 'Rebuild'
            else:
                print('(r)ebuild, rebuild (a)ll, (s)kip, (p)rint, (q)uit?: ', end='', flush=True)
                while True:
                    ch = 'r' if rebuildStale else getch()
                    if ch == 'q':
                        finish(0)
                    elif ch == 'p':
                        printCommands(commands)
                        checkResult = 'Printed'
                        break
                    elif ch == 's':
                        checkResult = 'Skip'
                        break
                    elif ch == 'r':
                        checkResult = 'Rebuild'
                        break
                    elif ch == 'a':
                        checkResult = 'Rebuild'
                        rebuildStale = True
                        break
        if checkResult == 'Printed':
            continue
        if checkResult == 'Skip':
            print('SKIPPING')
            continue
        clearCacheKey(stage)
        print('BUILDING:')
        os.chdir(stage['directory'])
        if not run(commands):
            print(prefix + ': FAILED')
            finish(1)
        writeCacheKey(stage)

if customRunCommand:
    os.chdir(executePath)
    if len(runCommand) == 1 and runCommand[0] == 'shell':
        print('Preparing interactive mode..')
        if win:
            modifiedEnv['PROMPT'] = '(prepare) $P$G'
            subprocess.run("cmd.exe", shell=True, env=modifiedEnv)
        else:
            prompt = '(prepare) %~ %# '
            with tempfile.NamedTemporaryFile(mode='w', delete=False) as tmp_zshrc:
                tmp_zshrc.write(f'export PS1="{prompt}"\n')
                tmp_zshrc_path = tmp_zshrc.name
            subprocess.run(['zsh', '--rcs', tmp_zshrc_path], env=modifiedEnv)
            os.remove(tmp_zshrc_path)
    elif not run(' '.join(runCommand) + '\n'):
        print('FAILED :(')
        finish(1)
    finish(0)

stage('patches', """
    git clone https://github.com/desktop-app/patches.git
    cd patches
    git checkout 412f65c296
""")

stage('msys64', """
win:
    SET PATH_BACKUP_=%PATH%
    SET PATH=%ROOT_DIR%\\ThirdParty\\msys64\\usr\\bin;%PATH%

    SET CHERE_INVOKING=enabled_from_arguments
    SET MSYS2_PATH_TYPE=inherit

    powershell -Command "iwr -OutFile ./msys64.exe https://github.com/msys2/msys2-installer/releases/download/2024-05-07/msys2-base-x86_64-20240507.sfx.exe"
    msys64.exe
    del msys64.exe

    bash -c "pacman-key --init; pacman-key --populate; pacman -Syu --noconfirm"
    pacman -Syu --noconfirm mingw-w64-x86_64-perl mingw-w64-x86_64-nasm mingw-w64-x86_64-yasm mingw-w64-x86_64-ninja msys/make diffutils pkg-config

    SET PATH=%PATH_BACKUP_%
""", 'ThirdParty')

stage('python', """
version: """ + (subprocess.run(['python', '-V'], capture_output=True, text=True, env=modifiedEnv).stdout.strip().split()[-1] if win else '0') + """
win:
    python -m venv python
    python\\Scripts\\activate.bat
    pip install pywin32 six meson
    deactivate
""", 'ThirdParty')

stage('NuGet', """
win:
    mkdir NuGet
    powershell -Command "iwr -OutFile ./NuGet/nuget.exe https://dist.nuget.org/win-x86-commandline/latest/nuget.exe"
""", 'ThirdParty')

stage('jom', """
win:
    powershell -Command "iwr -OutFile ./jom.zip https://master.qt.io/official_releases/jom/jom_1_1_3.zip"
    powershell -Command "Expand-Archive ./jom.zip"
    del jom.zip
""", 'ThirdParty')

stage('depot_tools', """
mac:
    git clone https://chromium.googlesource.com/chromium/tools/depot_tools.git
    cd depot_tools
    ./update_depot_tools
""", 'ThirdParty')

if not mac or 'build-stackwalk' in options:
    stage('gyp', """
win:
    git clone https://github.com/desktop-app/gyp.git
    cd gyp
    git checkout 618958fdbe
mac:
    python3 -m pip install \\
        --ignore-installed \\
        --target=$THIRDPARTY_DIR/gyp \\
        git+https://chromium.googlesource.com/external/gyp@master
""", 'ThirdParty')

stage('yasm', """
mac:
    git clone https://github.com/yasm/yasm.git
    cd yasm
    git checkout 41762bea
    ./autogen.sh
    make $MAKE_THREADS_CNT
""", 'ThirdParty')

stage('lzma', """
win:
    git clone https://github.com/desktop-app/lzma.git
    cd lzma\\C\\Util\\LzmaLib
    msbuild -m LzmaLib.sln /property:Configuration=Debug /property:Platform="$X8664"
release:
    msbuild -m LzmaLib.sln /property:Configuration=Release /property:Platform="$X8664"
""")

stage('xz', """
!win:
    git clone -b v5.4.5 https://github.com/tukaani-project/xz.git
    cd xz
    sed -i '' '\\@check_symbol_exists(futimens "sys/types.h;sys/stat.h" HAVE_FUTIMENS)@d' CMakeLists.txt
    CFLAGS="$UNGUARDED" CPPFLAGS="$UNGUARDED" cmake -B build . \\
        -D CMAKE_OSX_DEPLOYMENT_TARGET:STRING=$MACOSX_DEPLOYMENT_TARGET \\
        -D CMAKE_OSX_ARCHITECTURES="x86_64;arm64" \\
        -D CMAKE_INSTALL_PREFIX:STRING=$USED_PREFIX
    cmake --build build $MAKE_THREADS_CNT
    cmake --install build
""")

stage('zlib', """
    git clone https://github.com/madler/zlib.git
    cd zlib
    git checkout 643e17b749
win:
    cmake . ^
        -A %WIN32X64% ^
        -DCMAKE_MSVC_RUNTIME_LIBRARY="MultiThreaded$<$<CONFIG:Debug>:Debug>" ^
        -DCMAKE_C_FLAGS_DEBUG="/MTd /Zi /Ob0 /Od /RTC1" ^
        -DCMAKE_C_FLAGS_RELEASE="/MT /O2 /Ob2 /DNDEBUG" ^
        -DCMAKE_C_FLAGS="/DZLIB_WINAPI"
    cmake --build . --config Debug --parallel
release:
    cmake --build . --config Release --parallel
mac:
    CFLAGS="$MIN_VER $UNGUARDED" LDFLAGS="$MIN_VER" ./configure \\
        --static \\
        --prefix=$USED_PREFIX \\
        --archs="-arch x86_64 -arch arm64"
    make $MAKE_THREADS_CNT
    make install
""")

stage('mozjpeg', """
    git clone -b v4.1.5 https://github.com/mozilla/mozjpeg.git
    cd mozjpeg
win:
    cmake . ^
        -A %WIN32X64% ^
        -DWITH_JPEG8=ON ^
        -DPNG_SUPPORTED=OFF
    cmake --build . --config Debug --parallel
release:
    cmake --build . --config Release --parallel
mac:
    CFLAGS="-arch arm64" cmake -B build.arm64 . \\
        -D CMAKE_SYSTEM_NAME=Darwin \\
        -D CMAKE_SYSTEM_PROCESSOR=arm64 \\
        -D CMAKE_BUILD_TYPE=Release \\
        -D CMAKE_INSTALL_PREFIX=$USED_PREFIX \\
        -D CMAKE_OSX_DEPLOYMENT_TARGET:STRING=$MACOSX_DEPLOYMENT_TARGET \\
        -D WITH_JPEG8=ON \\
        -D ENABLE_SHARED=OFF \\
        -D PNG_SUPPORTED=OFF
    cmake --build build.arm64 $MAKE_THREADS_CNT
    CFLAGS="-arch x86_64" cmake -B build . \\
        -D CMAKE_SYSTEM_NAME=Darwin \\
        -D CMAKE_SYSTEM_PROCESSOR=x86_64 \\
        -D CMAKE_BUILD_TYPE=Release \\
        -D CMAKE_INSTALL_PREFIX=$USED_PREFIX \\
        -D CMAKE_OSX_DEPLOYMENT_TARGET:STRING=$MACOSX_DEPLOYMENT_TARGET \\
        -D WITH_JPEG8=ON \\
        -D ENABLE_SHARED=OFF \\
        -D PNG_SUPPORTED=OFF
    cmake --build build $MAKE_THREADS_CNT
    lipo -create build.arm64/libjpeg.a build/libjpeg.a -output build/libjpeg.a
    lipo -create build.arm64/libturbojpeg.a build/libturbojpeg.a -output build/libturbojpeg.a
    cmake --install build
""")

stage('openssl3', """
    git clone -b openssl-3.2.1 https://github.com/openssl/openssl openssl3
    cd openssl3
win32:
    perl Configure no-shared no-tests debug-VC-WIN32 /FS
win64:
    perl Configure no-shared no-tests debug-VC-WIN64A /FS
winarm:
    perl Configure no-shared no-tests debug-VC-WIN64-ARM /FS
win:
    jom -j%NUMBER_OF_PROCESSORS% build_libs
    mkdir out.dbg
    move libcrypto.lib out.dbg
    move libssl.lib out.dbg
    move ossl_static.pdb out.dbg
release:
    move out.dbg\\ossl_static.pdb out.dbg\\ossl_static
    jom clean
    move out.dbg\\ossl_static out.dbg\\ossl_static.pdb
win32_release:
    perl Configure no-shared no-tests VC-WIN32 /FS
win64_release:
    perl Configure no-shared no-tests VC-WIN64A /FS
winarm_release:
    perl Configure no-shared no-tests VC-WIN64-ARM /FS
win_release:
    jom -j%NUMBER_OF_PROCESSORS% build_libs
    mkdir out
    move libcrypto.lib out
    move libssl.lib out
    move ossl_static.pdb out
mac:
    ./Configure --prefix=$USED_PREFIX no-shared no-tests darwin64-arm64-cc $MIN_VER
    make build_libs $MAKE_THREADS_CNT
    mkdir out.arm64
    mv libssl.a out.arm64
    mv libcrypto.a out.arm64
    make clean
    ./Configure --prefix=$USED_PREFIX no-shared no-tests darwin64-x86_64-cc $MIN_VER
    make build_libs $MAKE_THREADS_CNT
    mkdir out.x86_64
    mv libssl.a out.x86_64
    mv libcrypto.a out.x86_64
    lipo -create out.arm64/libcrypto.a out.x86_64/libcrypto.a -output libcrypto.a
    lipo -create out.arm64/libssl.a out.x86_64/libssl.a -output libssl.a
""")

stage('opus', """
    git clone -b v1.3.1 https://github.com/xiph/opus.git
    cd opus
    git cherry-pick 927de8453c
win:
    cmake -B out . ^
        -A %WIN32X64% ^
        -DCMAKE_INSTALL_PREFIX=%LIBS_DIR%/local ^
        -DCMAKE_MSVC_RUNTIME_LIBRARY="MultiThreaded$<$<CONFIG:Debug>:Debug>" ^
        -DCMAKE_C_FLAGS_DEBUG="/MTd /Zi /Ob0 /Od /RTC1" ^
        -DCMAKE_C_FLAGS_RELEASE="/MT /O2 /Ob2 /DNDEBUG"
    cmake --build out --config Debug --parallel
    cmake --build out --config Release --parallel
    cmake --install out --config Release
mac:
    CFLAGS="$UNGUARDED" CPPFLAGS="$UNGUARDED" cmake -B build . \\
        -D CMAKE_OSX_DEPLOYMENT_TARGET:STRING=$MACOSX_DEPLOYMENT_TARGET \\
        -D CMAKE_OSX_ARCHITECTURES="x86_64;arm64" \\
        -D CMAKE_INSTALL_PREFIX:STRING=$USED_PREFIX
    cmake --build build $MAKE_THREADS_CNT
    cmake --install build
""")

stage('rnnoise', """
    git clone https://github.com/desktop-app/rnnoise.git
    cd rnnoise
    git checkout fe37e57d09
    mkdir out
    cd out
win:
    cmake -A %WIN32X64% ..
    cmake --build . --config Debug --parallel
release:
    cmake --build . --config Release --parallel
!win:
    mkdir Debug
    cd Debug
    cmake -G Ninja ../.. \\
        -D CMAKE_BUILD_TYPE=Debug \\
        -D CMAKE_OSX_ARCHITECTURES="x86_64;arm64"
    ninja
release:
    cd ..
    mkdir Release
    cd Release
    cmake -G Ninja ../.. \\
        -D CMAKE_BUILD_TYPE=Release \\
        -D CMAKE_OSX_ARCHITECTURES="x86_64;arm64"
    ninja
""")

stage('libiconv', """
mac:
    VERSION=1.17
    rm -f libiconv.tar.gz
    wget -O libiconv.tar.gz ftp://ftp.gnu.org/gnu/libiconv/libiconv-$VERSION.tar.gz
    rm -rf libiconv-$VERSION
    tar -xvzf libiconv.tar.gz
    rm libiconv.tar.gz
    mv libiconv-$VERSION libiconv
    cd libiconv
    CFLAGS="$MIN_VER $UNGUARDED -arch arm64" CPPFLAGS="$MIN_VER $UNGUARDED -arch arm64" LDFLAGS="$MIN_VER" ./configure --enable-static --host=arm --prefix=$USED_PREFIX
    make $MAKE_THREADS_CNT
    mkdir out.arm64
    mv lib/.libs/libiconv.a out.arm64
    make clean
    CFLAGS="$MIN_VER $UNGUARDED -arch x86_64" CPPFLAGS="$MIN_VER $UNGUARDED -arch x86_64" LDFLAGS="$MIN_VER" ./configure --enable-static --host=x86_64 --prefix=$USED_PREFIX
    make $MAKE_THREADS_CNT
    mkdir out.x86_64
    mv lib/.libs/libiconv.a out.x86_64
    lipo -create out.arm64/libiconv.a out.x86_64/libiconv.a -output lib/.libs/libiconv.a
    make install
""")

stage('gas-preprocessor', """
win:
    git clone https://github.com/FFmpeg/gas-preprocessor
    cd gas-preprocessor
    echo @echo off > cpp.bat
    echo cl %%%%%%** >> cpp.bat
""")

# Somehow in x86 Debug build dav1d crashes on AV1 10bpc videos.
stage('dav1d', """
    git clone -b 1.4.1 https://code.videolan.org/videolan/dav1d.git
    cd dav1d
win32:
    SET "TARGET=x86"
    SET "DAV1D_ASM_DISABLE=-Denable_asm=false"
win64:
    SET "TARGET=x86_64"
    SET "DAV1D_ASM_DISABLE="
winarm:
    SET "TARGET=aarch64"
    SET "DAV1D_ASM_DISABLE="
    SET "PATH_BACKUP_=%PATH%"
    SET "PATH=%LIBS_DIR%\\gas-preprocessor;%PATH%"
    echo armasm64 fails with 'syntax error in expression: tbnz x14, #4, 8f' as if this instruction is unknown/unsupported.
    git revert --no-edit d503bb0ccaf104b2f13da0f092e09cc9411b3297
win:
    set FILE=cross-file.txt
    echo [binaries] > %FILE%
    echo c = 'cl' >> %FILE%
    echo cpp = 'cl' >> %FILE%
    echo ar = 'lib' >> %FILE%
    echo windres = 'rc' >> %FILE%
    echo [host_machine] >> %FILE%
    echo system = 'windows' >> %FILE%
    echo cpu_family = '%TARGET%' >> %FILE%
    echo cpu = '%TARGET%' >> %FILE%
    echo endian = 'little' >> %FILE%

depends:python/Scripts/activate.bat
    %THIRDPARTY_DIR%\\python\\Scripts\\activate.bat
    meson setup --cross-file %FILE% --prefix %LIBS_DIR%/local --default-library=static --buildtype=debug -Denable_tools=false -Denable_tests=false %DAV1D_ASM_DISABLE% -Db_vscrt=mtd builddir-debug
    meson compile -C builddir-debug
    meson install -C builddir-debug
release:
    meson setup --cross-file %FILE% --prefix %LIBS_DIR%/local --default-library=static --buildtype=release -Denable_tools=false -Denable_tests=false -Db_vscrt=mt builddir-release
    meson compile -C builddir-release
    meson install -C builddir-release
win:
    copy %LIBS_DIR%\\local\\lib\\libdav1d.a %LIBS_DIR%\\local\\lib\\dav1d.lib
    deactivate
winarm:
    SET "PATH=%PATH_BACKUP_%"
mac:
    buildOneArch() {
        arch=$1
        folder=`pwd`/$2

        meson setup \\
            --cross-file ../patches/macos_meson_${arch}.txt \\
            --prefix ${USED_PREFIX} \\
            --default-library=static \\
            --buildtype=minsize \\
            -Denable_tools=false \\
            -Denable_tests=false \\
            ${folder}
        meson compile -C ${folder}
        meson install -C ${folder}

        mv ${USED_PREFIX}/lib/libdav1d.a ${folder}/libdav1d.a
    }

    buildOneArch arm64 build.arm64
    buildOneArch x86_64 build

    lipo -create build.arm64/libdav1d.a build/libdav1d.a -output ${USED_PREFIX}/lib/libdav1d.a
""")

stage('openh264', """
    git clone -b v2.4.1 https://github.com/cisco/openh264.git
    cd openh264
win32:
    SET "TARGET=x86"
win64:
    SET "TARGET=x86_64"
winarm:
    SET "TARGET=aarch64"
    SET "PATH_BACKUP_=%PATH%"
    SET "PATH=%LIBS_DIR%\\gas-preprocessor;%PATH%"
win:
    set FILE=cross-file.txt
    echo [binaries] > %FILE%
    echo c = 'cl' >> %FILE%
    echo cpp = 'cl' >> %FILE%
    echo ar = 'lib' >> %FILE%
    echo windres = 'rc' >> %FILE%
    echo [host_machine] >> %FILE%
    echo system = 'windows' >> %FILE%
    echo cpu_family = '%TARGET%' >> %FILE%
    echo cpu = '%TARGET%' >> %FILE%
    echo endian = 'little' >> %FILE%

depends:python/Scripts/activate.bat
    %THIRDPARTY_DIR%\\python\\Scripts\\activate.bat
    meson setup --cross-file %FILE% --prefix %LIBS_DIR%/local --default-library=static --buildtype=debug -Db_vscrt=mtd builddir-debug
    meson compile -C builddir-debug
    meson install -C builddir-debug
release:
    meson setup --cross-file %FILE% --prefix %LIBS_DIR%/local --default-library=static --buildtype=release -Db_vscrt=mt builddir-release
    meson compile -C builddir-release
    meson install -C builddir-release
win:
    copy %LIBS_DIR%\\local\\lib\\libopenh264.a %LIBS_DIR%\\local\\lib\\openh264.lib
    deactivate
winarm:
    SET "PATH=%PATH_BACKUP_%"
mac:
    buildOneArch() {
        arch=$1
        folder=`pwd`/$2

        meson setup \
            --cross-file ../patches/macos_meson_${arch}.txt \
            --prefix ${USED_PREFIX} \
            --default-library=static \
            --buildtype=minsize \
            ${folder}
        meson compile -C ${folder}
        meson install -C ${folder}

        mv ${USED_PREFIX}/lib/libopenh264.a ${folder}/libopenh264.a
    }

    buildOneArch aarch64 build.aarch64
    buildOneArch x86_64 build.x86_64

    lipo -create build.aarch64/libopenh264.a build.x86_64/libopenh264.a -output ${USED_PREFIX}/lib/libopenh264.a
""")

stage('libavif', """
    git clone -b v1.0.4 https://github.com/AOMediaCodec/libavif.git
    cd libavif
win:
    cmake . ^
        -A %WIN32X64% ^
        -DCMAKE_INSTALL_PREFIX=%LIBS_DIR%/local ^
        -DCMAKE_MSVC_RUNTIME_LIBRARY="MultiThreaded$<$<CONFIG:Debug>:Debug>" ^
        -DCMAKE_C_FLAGS_DEBUG="/MTd /Zi /Ob0 /Od /RTC1" ^
        -DCMAKE_C_FLAGS_RELEASE="/MT /O2 /Ob2 /DNDEBUG" ^
        -DBUILD_SHARED_LIBS=OFF ^
        -DAVIF_ENABLE_WERROR=OFF ^
        -DAVIF_CODEC_DAV1D=ON
    cmake --build . --config Debug --parallel
    cmake --install . --config Debug
release:
    cmake --build . --config Release --parallel
    cmake --install . --config Release
mac:
    cmake . \\
        -D CMAKE_OSX_ARCHITECTURES="x86_64;arm64" \\
        -D CMAKE_OSX_DEPLOYMENT_TARGET:STRING=$MACOSX_DEPLOYMENT_TARGET \\
        -D CMAKE_INSTALL_PREFIX:STRING=$USED_PREFIX \\
        -D BUILD_SHARED_LIBS=OFF \\
        -D AVIF_ENABLE_WERROR=OFF \\
        -D AVIF_CODEC_DAV1D=ON \\
        -D CMAKE_DISABLE_FIND_PACKAGE_libsharpyuv=ON
    cmake --build . --config MinSizeRel $MAKE_THREADS_CNT
    cmake --install . --config MinSizeRel
""")

stage('libde265', """
    git clone -b v1.0.15 https://github.com/strukturag/libde265.git
    cd libde265
win:
    cmake . ^
        -A %WIN32X64% ^
        -DCMAKE_INSTALL_PREFIX=%LIBS_DIR%/local ^
        -DCMAKE_MSVC_RUNTIME_LIBRARY="MultiThreaded$<$<CONFIG:Debug>:Debug>" ^
        -DCMAKE_C_FLAGS="/DLIBDE265_STATIC_BUILD" ^
        -DCMAKE_CXX_FLAGS="/DLIBDE265_STATIC_BUILD" ^
        -DCMAKE_C_FLAGS_DEBUG="/MTd /Zi /Ob0 /Od /RTC1" ^
        -DCMAKE_CXX_FLAGS_DEBUG="/MTd /Zi /Ob0 /Od /RTC1" ^
        -DCMAKE_C_FLAGS_RELEASE="/MT /O2 /Ob2 /DNDEBUG" ^
        -DCMAKE_CXX_FLAGS_RELEASE="/MT /O2 /Ob2 /DNDEBUG" ^
        -DENABLE_SDL=OFF ^
        -DBUILD_SHARED_LIBS=OFF ^
        -DENABLE_DECODER=OFF ^
        -DENABLE_ENCODER=OFF
    cmake --build . --config Debug --parallel
    cmake --install . --config Debug
release:
    cmake --build . --config Release --parallel
    cmake --install . --config Release
mac:
    cmake . \\
        -D CMAKE_OSX_ARCHITECTURES="x86_64;arm64" \\
        -D CMAKE_OSX_DEPLOYMENT_TARGET:STRING=$MACOSX_DEPLOYMENT_TARGET \\
        -D CMAKE_INSTALL_PREFIX:STRING=$USED_PREFIX \\
        -D DISABLE_SSE=ON \\
        -D ENABLE_SDL=OFF \\
        -D BUILD_SHARED_LIBS=OFF \\
        -D ENABLE_DECODER=ON \\
        -D ENABLE_ENCODER=OFF
    cmake --build . --config MinSizeRel $MAKE_THREADS_CNT
    cmake --install . --config MinSizeRel
""")

stage('libwebp', """
    git clone -b v1.4.0 https://github.com/webmproject/libwebp.git
    cd libwebp
win:
    nmake /f Makefile.vc CFG=debug-static OBJDIR=out RTLIBCFG=static all
    nmake /f Makefile.vc CFG=release-static OBJDIR=out RTLIBCFG=static all
    copy out\\release-static\\$X8664\\lib\\libwebp.lib out\\release-static\\$X8664\\lib\\webp.lib
    copy out\\release-static\\$X8664\\lib\\libwebpdemux.lib out\\release-static\\$X8664\\lib\\webpdemux.lib
    copy out\\release-static\\$X8664\\lib\\libwebpmux.lib out\\release-static\\$X8664\\lib\\webpmux.lib
mac:
    buildOneArch() {
        arch=$1
        folder=$2

        CFLAGS=$UNGUARDED cmake -B $folder -G Ninja . \\
            -D CMAKE_BUILD_TYPE=Release \\
            -D CMAKE_INSTALL_PREFIX=$USED_PREFIX \\
            -D CMAKE_OSX_DEPLOYMENT_TARGET:STRING=$MACOSX_DEPLOYMENT_TARGET \\
            -D CMAKE_OSX_ARCHITECTURES=$arch \\
            -D WEBP_BUILD_ANIM_UTILS=OFF \\
            -D WEBP_BUILD_CWEBP=OFF \\
            -D WEBP_BUILD_DWEBP=OFF \\
            -D WEBP_BUILD_GIF2WEBP=OFF \\
            -D WEBP_BUILD_IMG2WEBP=OFF \\
            -D WEBP_BUILD_VWEBP=OFF \\
            -D WEBP_BUILD_WEBPMUX=OFF \\
            -D WEBP_BUILD_WEBPINFO=OFF \\
            -D WEBP_BUILD_EXTRAS=OFF
        cmake --build $folder $MAKE_THREADS_CNT
    }
    buildOneArch arm64 build.arm64
    buildOneArch x86_64 build

    lipo -create build.arm64/libsharpyuv.a build/libsharpyuv.a -output build/libsharpyuv.a
    lipo -create build.arm64/libwebp.a build/libwebp.a -output build/libwebp.a
    lipo -create build.arm64/libwebpdemux.a build/libwebpdemux.a -output build/libwebpdemux.a
    lipo -create build.arm64/libwebpmux.a build/libwebpmux.a -output build/libwebpmux.a
    cmake --install build
""")

stage('libheif', """
    git clone -b v1.17.6 https://github.com/strukturag/libheif.git
    cd libheif
win:
    %THIRDPARTY_DIR%\\msys64\\usr\\bin\\sed.exe -i 's/LIBHEIF_EXPORTS/LIBDE265_STATIC_BUILD/g' libheif/CMakeLists.txt
    %THIRDPARTY_DIR%\\msys64\\usr\\bin\\sed.exe -i 's/HAVE_VISIBILITY/LIBHEIF_STATIC_BUILD/g' libheif/CMakeLists.txt
    cmake . ^
        -A %WIN32X64% ^
        -DCMAKE_INSTALL_PREFIX=%LIBS_DIR%/local ^
        -DCMAKE_MSVC_RUNTIME_LIBRARY="MultiThreaded$<$<CONFIG:Debug>:Debug>" ^
        -DCMAKE_C_FLAGS_DEBUG="/MTd /Zi /Ob0 /Od /RTC1" ^
        -DCMAKE_CXX_FLAGS_DEBUG="/MTd /Zi /Ob0 /Od /RTC1" ^
        -DCMAKE_C_FLAGS_RELEASE="/MT /O2 /Ob2 /DNDEBUG" ^
        -DCMAKE_CXX_FLAGS_RELEASE="/MT /O2 /Ob2 /DNDEBUG" ^
        -DBUILD_SHARED_LIBS=OFF ^
        -DBUILD_TESTING=OFF ^
        -DENABLE_PLUGIN_LOADING=OFF ^
        -DWITH_LIBDE265=ON ^
        -DWITH_SvtEnc=OFF ^
        -DWITH_SvtEnc_PLUGIN=OFF ^
        -DWITH_RAV1E=OFF ^
        -DWITH_RAV1E_PLUGIN=OFF ^
        -DWITH_EXAMPLES=OFF
    cmake --build . --config Debug --parallel
    cmake --install . --config Debug
release:
    cmake --build . --config Release --parallel
    cmake --install . --config Release
mac:
    cmake . \\
        -D CMAKE_OSX_ARCHITECTURES="x86_64;arm64" \\
        -D CMAKE_OSX_DEPLOYMENT_TARGET:STRING=$MACOSX_DEPLOYMENT_TARGET \\
        -D CMAKE_INSTALL_PREFIX:STRING=$USED_PREFIX \\
        -D BUILD_SHARED_LIBS=OFF \\
        -D BUILD_TESTING=OFF \\
        -D ENABLE_PLUGIN_LOADING=OFF \\
        -D WITH_AOM_ENCODER=OFF \\
        -D WITH_AOM_DECODER=OFF \\
        -D WITH_X265=OFF \\
        -D WITH_SvtEnc=OFF \\
        -D WITH_RAV1E=OFF \\
        -D WITH_DAV1D=ON \\
        -D WITH_LIBDE265=ON \\
        -D LIBDE265_INCLUDE_DIR=$USED_PREFIX/include/ \\
        -D LIBDE265_LIBRARY=$USED_PREFIX/lib/libde265.a \\
        -D LIBSHARPYUV_INCLUDE_DIR=$USED_PREFIX/include/webp/ \\
        -D LIBSHARPYUV_LIBRARY=$USED_PREFIX/lib/libsharpyuv.a \\
        -D WITH_EXAMPLES=OFF
    cmake --build . --config MinSizeRel $MAKE_THREADS_CNT
    cmake --install . --config MinSizeRel
""")

stage('libjxl', """
    git clone -b v0.10.3 --recursive --shallow-submodules https://github.com/libjxl/libjxl.git
    cd libjxl
""" + setVar("cmake_defines", """
    -DBUILD_SHARED_LIBS=OFF
    -DBUILD_TESTING=OFF
    -DJPEGXL_ENABLE_FUZZERS=OFF
    -DJPEGXL_ENABLE_DEVTOOLS=OFF
    -DJPEGXL_ENABLE_TOOLS=OFF
    -DJPEGXL_ENABLE_DOXYGEN=OFF
    -DJPEGXL_ENABLE_MANPAGES=OFF
    -DJPEGXL_ENABLE_EXAMPLES=OFF
    -DJPEGXL_ENABLE_JNI=OFF
    -DJPEGXL_ENABLE_JPEGLI_LIBJPEG=OFF
    -DJPEGXL_ENABLE_SJPEG=OFF
    -DJPEGXL_ENABLE_OPENEXR=OFF
    -DJPEGXL_ENABLE_SKCMS=ON
    -DJPEGXL_ENABLE_VIEWERS=OFF
    -DJPEGXL_ENABLE_TCMALLOC=OFF
    -DJPEGXL_ENABLE_PLUGINS=OFF
    -DJPEGXL_ENABLE_COVERAGE=OFF
    -DJPEGXL_WARNINGS_AS_ERRORS=OFF
""") + """
win:
    cmake . ^
        -A %WIN32X64% ^
        -DCMAKE_INSTALL_PREFIX=%LIBS_DIR%/local ^
        -DCMAKE_MSVC_RUNTIME_LIBRARY="MultiThreaded$<$<CONFIG:Debug>:Debug>" ^
        -DCMAKE_C_FLAGS="/DJXL_STATIC_DEFINE /DJXL_THREADS_STATIC_DEFINE /DJXL_CMS_STATIC_DEFINE" ^
        -DCMAKE_CXX_FLAGS="/DJXL_STATIC_DEFINE /DJXL_THREADS_STATIC_DEFINE /DJXL_CMS_STATIC_DEFINE" ^
        -DCMAKE_C_FLAGS_DEBUG="/MTd /Zi /Ob0 /Od /RTC1" ^
        -DCMAKE_CXX_FLAGS_DEBUG="/MTd /Zi /Ob0 /Od /RTC1" ^
        -DCMAKE_C_FLAGS_RELEASE="/MT /O2 /Ob2 /DNDEBUG" ^
        -DCMAKE_CXX_FLAGS_RELEASE="/MT /O2 /Ob2 /DNDEBUG" ^
        %cmake_defines%
    cmake --build . --config Debug --parallel
    cmake --install . --config Debug
release:
    cmake --build . --config Release --parallel
    cmake --install . --config Release
mac:
    cmake . \\
        -D CMAKE_OSX_ARCHITECTURES="x86_64;arm64" \\
        -D CMAKE_OSX_DEPLOYMENT_TARGET:STRING=$MACOSX_DEPLOYMENT_TARGET \\
        -D CMAKE_INSTALL_PREFIX:STRING=$USED_PREFIX \\
        ${cmake_defines}
    cmake --build . --config MinSizeRel $MAKE_THREADS_CNT
    cmake --install . --config MinSizeRel
""")

stage('libvpx', """
    git clone https://github.com/webmproject/libvpx.git
depends:patches/libvpx/*.patch
    cd libvpx
    git checkout v1.14.1
win:
    for /r %%i in (..\\patches\\libvpx\\*) do git apply %%i

    SET PATH_BACKUP_=%PATH%
    SET PATH=%ROOT_DIR%\\ThirdParty\\msys64\\usr\\bin;%PATH%

    SET CHERE_INVOKING=enabled_from_arguments
    SET MSYS2_PATH_TYPE=inherit

win32:
    SET "TOOLCHAIN=x86-win32-vs17"
win64:
    SET "TOOLCHAIN=x86_64-win64-vs17"
winarm:
    SET "TOOLCHAIN=arm64-win64-vs17"
win:
depends:patches/build_libvpx_win.sh
    bash --login ../patches/build_libvpx_win.sh

    SET PATH=%PATH_BACKUP_%
mac:
    find ../patches/libvpx -type f -print0 | sort -z | xargs -0 git apply

depends:yasm/yasm
    ./configure --prefix=$USED_PREFIX \
    --target=arm64-darwin20-gcc \
    --disable-examples \
    --disable-unit-tests \
    --disable-tools \
    --disable-docs \
    --enable-vp8 \
    --enable-vp9 \
    --enable-webm-io \
    --size-limit=4096x4096

    make $MAKE_THREADS_CNT

    mkdir out.arm64
    mv libvpx.a out.arm64

    make clean

    ./configure --prefix=$USED_PREFIX \
    --target=x86_64-darwin20-gcc \
    --disable-examples \
    --disable-unit-tests \
    --disable-tools \
    --disable-docs \
    --enable-vp8 \
    --enable-vp9 \
    --enable-webm-io

    make $MAKE_THREADS_CNT

    mkdir out.x86_64
    mv libvpx.a out.x86_64

    lipo -create out.arm64/libvpx.a out.x86_64/libvpx.a -output libvpx.a

    make install
""")

stage('liblcms2', """
mac:
    git clone -b lcms2.16 https://github.com/mm2/Little-CMS.git liblcms2
    cd liblcms2

    buildOneArch() {
        arch=$1
        folder=`pwd`/$2

        meson setup \\
            --cross-file ../patches/macos_meson_${arch}.txt \\
            --prefix ${USED_PREFIX} \\
            --default-library=static \\
            --buildtype=minsize \\
            ${folder}
        meson compile -C ${folder}
        meson install -C ${folder}

        mv ${USED_PREFIX}/lib/liblcms2.a ${folder}/liblcms2.a
    }

    buildOneArch arm64 build.arm64
    buildOneArch x86_64 build

    lipo -create build.arm64/liblcms2.a build/liblcms2.a -output ${USED_PREFIX}/lib/liblcms2.a
""")

stage('nv-codec-headers', """
win:
    git clone -b n12.1.14.0 https://github.com/FFmpeg/nv-codec-headers.git
""")

stage('regex', """
    git clone -b boost-1.83.0 https://github.com/boostorg/regex.git
""")

stage('ffmpeg', """
    git clone -b n6.1.1 https://github.com/FFmpeg/FFmpeg.git ffmpeg
    cd ffmpeg
win:
depends:patches/ffmpeg.patch
    git apply ../patches/ffmpeg.patch

    SET PATH_BACKUP_=%PATH%
    SET PATH=%ROOT_DIR%\\ThirdParty\\msys64\\usr\\bin;%PATH%

    SET CHERE_INVOKING=enabled_from_arguments
    SET MSYS2_PATH_TYPE=inherit

    SET "ARCH_PARAM="
winarm:
    SET "ARCH_PARAM=--arch=aarch64"
win:
depends:patches/build_ffmpeg_win.sh
    bash --login ../patches/build_ffmpeg_win.sh

    SET PATH=%PATH_BACKUP_%
mac:
    export PKG_CONFIG_PATH=$USED_PREFIX/lib/pkgconfig
depends:yasm/yasm

    configureFFmpeg() {
        arch=$1

        ./configure --prefix=$USED_PREFIX \
        --enable-cross-compile \
        --target-os=darwin \
        --arch="$arch" \
        --extra-cflags="$MIN_VER -arch $arch $UNGUARDED -DCONFIG_SAFE_BITSTREAM_READER=1 -I$USED_PREFIX/include" \
        --extra-cxxflags="$MIN_VER -arch $arch $UNGUARDED -DCONFIG_SAFE_BITSTREAM_READER=1 -I$USED_PREFIX/include" \
        --extra-ldflags="$MIN_VER -arch $arch $USED_PREFIX/lib/libopus.a -lc++" \
        --disable-programs \
        --disable-doc \
        --disable-network \
        --disable-everything \
        --enable-protocol=file \
        --enable-libdav1d \
        --enable-libopenh264 \
        --enable-libopus \
        --enable-libvpx \
        --enable-hwaccel=h264_videotoolbox \
        --enable-hwaccel=hevc_videotoolbox \
        --enable-hwaccel=mpeg1_videotoolbox \
        --enable-hwaccel=mpeg2_videotoolbox \
        --enable-hwaccel=mpeg4_videotoolbox \
        --enable-decoder=aac \
        --enable-decoder=aac_at \
        --enable-decoder=aac_fixed \
        --enable-decoder=aac_latm \
        --enable-decoder=aasc \
        --enable-decoder=ac3 \
        --enable-decoder=alac \
        --enable-decoder=alac_at \
        --enable-decoder=av1 \
        --enable-decoder=eac3 \
        --enable-decoder=flac \
        --enable-decoder=gif \
        --enable-decoder=h264 \
        --enable-decoder=hevc \
        --enable-decoder=libdav1d \
        --enable-decoder=libvpx_vp8 \
        --enable-decoder=libvpx_vp9 \
        --enable-decoder=mp1 \
        --enable-decoder=mp1float \
        --enable-decoder=mp2 \
        --enable-decoder=mp2float \
        --enable-decoder=mp3 \
        --enable-decoder=mp3adu \
        --enable-decoder=mp3adufloat \
        --enable-decoder=mp3float \
        --enable-decoder=mp3on4 \
        --enable-decoder=mp3on4float \
        --enable-decoder=mpeg4 \
        --enable-decoder=msmpeg4v2 \
        --enable-decoder=msmpeg4v3 \
        --enable-decoder=opus \
        --enable-decoder=pcm_alaw \
        --enable-decoder=pcm_alaw_at \
        --enable-decoder=pcm_f32be \
        --enable-decoder=pcm_f32le \
        --enable-decoder=pcm_f64be \
        --enable-decoder=pcm_f64le \
        --enable-decoder=pcm_lxf \
        --enable-decoder=pcm_mulaw \
        --enable-decoder=pcm_mulaw_at \
        --enable-decoder=pcm_s16be \
        --enable-decoder=pcm_s16be_planar \
        --enable-decoder=pcm_s16le \
        --enable-decoder=pcm_s16le_planar \
        --enable-decoder=pcm_s24be \
        --enable-decoder=pcm_s24daud \
        --enable-decoder=pcm_s24le \
        --enable-decoder=pcm_s24le_planar \
        --enable-decoder=pcm_s32be \
        --enable-decoder=pcm_s32le \
        --enable-decoder=pcm_s32le_planar \
        --enable-decoder=pcm_s64be \
        --enable-decoder=pcm_s64le \
        --enable-decoder=pcm_s8 \
        --enable-decoder=pcm_s8_planar \
        --enable-decoder=pcm_u16be \
        --enable-decoder=pcm_u16le \
        --enable-decoder=pcm_u24be \
        --enable-decoder=pcm_u24le \
        --enable-decoder=pcm_u32be \
        --enable-decoder=pcm_u32le \
        --enable-decoder=pcm_u8 \
        --enable-decoder=vorbis \
        --enable-decoder=vp8 \
        --enable-decoder=wavpack \
        --enable-decoder=wmalossless \
        --enable-decoder=wmapro \
        --enable-decoder=wmav1 \
        --enable-decoder=wmav2 \
        --enable-decoder=wmavoice \
        --enable-encoder=aac \
        --enable-encoder=libopus \
        --enable-encoder=libopenh264 \
        --enable-filter=atempo \
        --enable-parser=aac \
        --enable-parser=aac_latm \
        --enable-parser=flac \
        --enable-parser=gif \
        --enable-parser=h264 \
        --enable-parser=hevc \
        --enable-parser=mpeg4video \
        --enable-parser=mpegaudio \
        --enable-parser=opus \
        --enable-parser=vorbis \
        --enable-demuxer=aac \
        --enable-demuxer=flac \
        --enable-demuxer=gif \
        --enable-demuxer=h264 \
        --enable-demuxer=hevc \
        --enable-demuxer=matroska \
        --enable-demuxer=m4v \
        --enable-demuxer=mov \
        --enable-demuxer=mp3 \
        --enable-demuxer=ogg \
        --enable-demuxer=wav \
        --enable-muxer=mp4 \
        --enable-muxer=ogg \
        --enable-muxer=opus
    }

    configureFFmpeg arm64
    make $MAKE_THREADS_CNT

    mkdir out.arm64
    mv libavfilter/libavfilter.a out.arm64
    mv libavformat/libavformat.a out.arm64
    mv libavcodec/libavcodec.a out.arm64
    mv libswresample/libswresample.a out.arm64
    mv libswscale/libswscale.a out.arm64
    mv libavutil/libavutil.a out.arm64

    make clean

    configureFFmpeg x86_64
    make $MAKE_THREADS_CNT

    mkdir out.x86_64
    mv libavfilter/libavfilter.a out.x86_64
    mv libavformat/libavformat.a out.x86_64
    mv libavcodec/libavcodec.a out.x86_64
    mv libswresample/libswresample.a out.x86_64
    mv libswscale/libswscale.a out.x86_64
    mv libavutil/libavutil.a out.x86_64

    lipo -create out.arm64/libavfilter.a out.x86_64/libavfilter.a -output libavfilter/libavfilter.a
    lipo -create out.arm64/libavformat.a out.x86_64/libavformat.a -output libavformat/libavformat.a
    lipo -create out.arm64/libavcodec.a out.x86_64/libavcodec.a -output libavcodec/libavcodec.a
    lipo -create out.arm64/libswresample.a out.x86_64/libswresample.a -output libswresample/libswresample.a
    lipo -create out.arm64/libswscale.a out.x86_64/libswscale.a -output libswscale/libswscale.a
    lipo -create out.arm64/libavutil.a out.x86_64/libavutil.a -output libavutil/libavutil.a

    make install
""")

stage('openal-soft', """
version: 3
win:
    git clone -b wasapi_exact_device_time https://github.com/telegramdesktop/openal-soft.git
    cd openal-soft
    cmake -B build . ^
        -A %WIN32X64% ^
        -D LIBTYPE:STRING=STATIC ^
        -D FORCE_STATIC_VCRT=ON
    cmake --build build --config Debug --parallel
release:
    cmake --build build --config RelWithDebInfo --parallel
mac:
    git clone -b coreaudio_device_uid https://github.com/telegramdesktop/openal-soft.git
    cd openal-soft
    CFLAGS=$UNGUARDED CPPFLAGS=$UNGUARDED cmake -B build . \\
        -D CMAKE_BUILD_TYPE=RelWithDebInfo \\
        -D CMAKE_INSTALL_PREFIX:PATH=$USED_PREFIX \\
        -D ALSOFT_EXAMPLES=OFF \\
        -D ALSOFT_UTILS=OFF \\
        -D LIBTYPE:STRING=STATIC \\
        -D CMAKE_OSX_DEPLOYMENT_TARGET:STRING=$MACOSX_DEPLOYMENT_TARGET \\
        -D CMAKE_OSX_ARCHITECTURES="x86_64;arm64"
    cmake --build build $MAKE_THREADS_CNT
    cmake --install build
""")

if 'build-stackwalk' in options:
    stage('stackwalk', """
mac:
    git clone https://chromium.googlesource.com/breakpad/breakpad stackwalk
    cd stackwalk
    git checkout dfcb7b6799
depends:patches/breakpad.diff
    git apply ../patches/breakpad.diff
    git clone -b release-1.11.0 https://github.com/google/googletest src/testing
    git clone https://chromium.googlesource.com/linux-syscall-support src/third_party/lss
    cd src/third_party/lss
    git checkout e1e7b0ad8e
    cd ../../build
    PYTHONPATH=$THIRDPARTY_DIR/gyp python3 gyp_breakpad
    cd ../processor
    xcodebuild -project processor.xcodeproj -target minidump_stackwalk -configuration Release build
""")

stage('breakpad', """
    git clone https://chromium.googlesource.com/breakpad/breakpad
    cd breakpad
    git checkout dfcb7b6799
depends:patches/breakpad.diff
    git apply ../patches/breakpad.diff
    git clone -b release-1.11.0 https://github.com/google/googletest src/testing
win:
    SET "PYTHONUTF8=1"
    SET "FolderPostfix="
win64:
    SET "FolderPostfix=_x64"
winarm:
    SET "FolderPostfix=_ARM64"
win:
depends:python/Scripts/activate.bat
    %THIRDPARTY_DIR%\\python\\Scripts\\activate.bat
    cd src\\client\\windows
    gyp --no-circular-check breakpad_client.gyp --format=ninja
    cd ..\\..
    ninja -C out/Debug%FolderPostfix% common crash_generation_client exception_handler
release:
    ninja -C out/Release%FolderPostfix% common crash_generation_client exception_handler
    cd tools\\windows\\dump_syms
    gyp dump_syms.gyp --format=msvs
    msbuild -m dump_syms.vcxproj /property:Configuration=Release /property:Platform="x64"
win:
    deactivate
mac:
    git clone https://chromium.googlesource.com/linux-syscall-support src/third_party/lss
    cd src/third_party/lss
    git checkout e1e7b0ad8e
    cd ../../..
    cd src/client/mac
    xcodebuild -project Breakpad.xcodeproj -target Breakpad -configuration Debug build
release:
    xcodebuild -project Breakpad.xcodeproj -target Breakpad -configuration Release build
    cd ../../tools/mac/dump_syms
    xcodebuild -project dump_syms.xcodeproj -target dump_syms -configuration Release build
""")

stage('crashpad', """
mac:
    git clone https://github.com/desktop-app/crashpad.git
    cd crashpad
    git checkout 3279fae3f0
    git submodule init
    git submodule update third_party/mini_chromium
    ZLIB_PATH=$USED_PREFIX/include
    ZLIB_LIB=$USED_PREFIX/lib/libz.a
    mkdir out
    cd out
    mkdir Debug.x86_64
    cd Debug.x86_64
    cmake -G Ninja \
        -DCMAKE_BUILD_TYPE=Debug \
        -DCMAKE_OSX_ARCHITECTURES=x86_64 \
        -DCRASHPAD_SPECIAL_TARGET=$SPECIAL_TARGET \
        -DCRASHPAD_ZLIB_INCLUDE_PATH=$ZLIB_PATH \
        -DCRASHPAD_ZLIB_LIB_PATH=$ZLIB_LIB ../..
    ninja
    cd ..
    mkdir Debug.arm64
    cd Debug.arm64
    cmake -G Ninja \
        -DCMAKE_BUILD_TYPE=Debug \
        -DCMAKE_OSX_ARCHITECTURES=arm64 \
        -DCRASHPAD_SPECIAL_TARGET=$SPECIAL_TARGET \
        -DCRASHPAD_ZLIB_INCLUDE_PATH=$ZLIB_PATH \
        -DCRASHPAD_ZLIB_LIB_PATH=$ZLIB_LIB ../..
    ninja
    cd ..
    mkdir Debug
    lipo -create Debug.arm64/crashpad_handler Debug.x86_64/crashpad_handler -output Debug/crashpad_handler
    lipo -create Debug.arm64/libcrashpad_client.a Debug.x86_64/libcrashpad_client.a -output Debug/libcrashpad_client.a
release:
    mkdir Release.x86_64
    cd Release.x86_64
    cmake -G Ninja \
        -DCMAKE_BUILD_TYPE=Release \
        -DCMAKE_OSX_ARCHITECTURES=x86_64 \
        -DCRASHPAD_SPECIAL_TARGET=$SPECIAL_TARGET \
        -DCRASHPAD_ZLIB_INCLUDE_PATH=$ZLIB_PATH \
        -DCRASHPAD_ZLIB_LIB_PATH=$ZLIB_LIB ../..
    ninja
    cd ..
    mkdir Release.arm64
    cd Release.arm64
    cmake -G Ninja \
        -DCMAKE_BUILD_TYPE=Release \
        -DCMAKE_OSX_ARCHITECTURES=arm64 \
        -DCRASHPAD_SPECIAL_TARGET=$SPECIAL_TARGET \
        -DCRASHPAD_ZLIB_INCLUDE_PATH=$ZLIB_PATH \
        -DCRASHPAD_ZLIB_LIB_PATH=$ZLIB_LIB ../..
    ninja
    cd ..
    mkdir Release
    lipo -create Release.arm64/crashpad_handler Release.x86_64/crashpad_handler -output Release/crashpad_handler
    lipo -create Release.arm64/libcrashpad_client.a Release.x86_64/libcrashpad_client.a -output Release/libcrashpad_client.a
""")

if qt < '6':
    if win:
        stage('tg_angle', """
win:
    git clone https://github.com/desktop-app/tg_angle.git
    cd tg_angle
    git checkout e3f59e8d0c
    mkdir out
    cd out
    mkdir Debug
    cd Debug
    cmake -G Ninja ^
        -DCMAKE_BUILD_TYPE=Debug ^
        -DTG_ANGLE_SPECIAL_TARGET=%SPECIAL_TARGET% ^
        -DTG_ANGLE_ZLIB_INCLUDE_PATH=%LIBS_DIR%/zlib ../..
    ninja
release:
    cd ..
    mkdir Release
    cd Release
    cmake -G Ninja ^
        -DCMAKE_BUILD_TYPE=Release ^
        -DTG_ANGLE_SPECIAL_TARGET=%SPECIAL_TARGET% ^
        -DTG_ANGLE_ZLIB_INCLUDE_PATH=%LIBS_DIR%/zlib ../..
    ninja
    cd ..\\..\\..
""")

    stage('qt_' + qt, """
    git clone -b v$QT-lts-lgpl https://github.com/qt/qt5.git qt_$QT
    cd qt_$QT
    git submodule update --init --recursive --progress qtbase qtimageformats qtsvg
depends:patches/qtbase_""" + qt + """/*.patch
    cd qtbase
win:
    git revert --no-edit 6ad56dce34
    setlocal enabledelayedexpansion
    for /r %%i in (..\\..\\patches\\qtbase_%QT%\\*) do (
        git apply %%i -v
        if errorlevel 1 (
            echo ERROR: Applying patch %%~nxi failed!
            exit /b 1
        )
    )

    cd ..

    SET CONFIGURATIONS=-debug
release:
    SET CONFIGURATIONS=-debug-and-release
win:
    """ + removeDir('"%LIBS_DIR%\\Qt-' + qt + '"') + """
    SET ANGLE_DIR=%LIBS_DIR%\\tg_angle
    SET ANGLE_LIBS_DIR=%ANGLE_DIR%\\out
    SET MOZJPEG_DIR=%LIBS_DIR%\\mozjpeg
    SET OPENSSL_DIR=%LIBS_DIR%\\openssl3
    SET OPENSSL_LIBS_DIR=%OPENSSL_DIR%\\out
    SET ZLIB_LIBS_DIR=%LIBS_DIR%\\zlib
    SET WEBP_DIR=%LIBS_DIR%\\libwebp
    configure -prefix "%LIBS_DIR%\\Qt-%QT%" ^
        %CONFIGURATIONS% ^
        -force-debug-info ^
        -opensource ^
        -confirm-license ^
        -static ^
        -static-runtime ^
        -opengl es2 -no-angle ^
        -I "%ANGLE_DIR%\\include" ^
        -D "KHRONOS_STATIC=" ^
        -D "DESKTOP_APP_QT_STATIC_ANGLE=" ^
        QMAKE_LIBS_OPENGL_ES2_DEBUG="%ANGLE_LIBS_DIR%\\Debug\\tg_angle.lib %ZLIB_LIBS_DIR%\\Debug\\zlibstaticd.lib d3d9.lib dxgi.lib dxguid.lib" ^
        QMAKE_LIBS_OPENGL_ES2_RELEASE="%ANGLE_LIBS_DIR%\\Release\\tg_angle.lib %ZLIB_LIBS_DIR%\\Release\\zlibstatic.lib d3d9.lib dxgi.lib dxguid.lib" ^
        -egl ^
        QMAKE_LIBS_EGL_DEBUG="%ANGLE_LIBS_DIR%\\Debug\\tg_angle.lib %ZLIB_LIBS_DIR%\\Debug\\zlibstaticd.lib d3d9.lib dxgi.lib dxguid.lib Gdi32.lib User32.lib" ^
        QMAKE_LIBS_EGL_RELEASE="%ANGLE_LIBS_DIR%\\Release\\tg_angle.lib %ZLIB_LIBS_DIR%\\Release\\zlibstatic.lib d3d9.lib dxgi.lib dxguid.lib Gdi32.lib User32.lib" ^
        -openssl-linked ^
        -I "%OPENSSL_DIR%\\include" ^
        OPENSSL_LIBS_DEBUG="%OPENSSL_LIBS_DIR%.dbg\\libssl.lib %OPENSSL_LIBS_DIR%.dbg\\libcrypto.lib Ws2_32.lib Gdi32.lib Advapi32.lib Crypt32.lib User32.lib" ^
        OPENSSL_LIBS_RELEASE="%OPENSSL_LIBS_DIR%\\libssl.lib %OPENSSL_LIBS_DIR%\\libcrypto.lib Ws2_32.lib Gdi32.lib Advapi32.lib Crypt32.lib User32.lib" ^
        -I "%MOZJPEG_DIR%" ^
        LIBJPEG_LIBS_DEBUG="%MOZJPEG_DIR%\\Debug\\jpeg-static.lib" ^
        LIBJPEG_LIBS_RELEASE="%MOZJPEG_DIR%\\Release\\jpeg-static.lib" ^
        -system-webp ^
        -I "%WEBP_DIR%\\src" ^
        -L "%WEBP_DIR%\\out\\release-static\\$X8664\\lib" ^
        -mp ^
        -no-feature-netlistmgr ^
        -nomake examples ^
        -nomake tests ^
        -platform win32-msvc

    jom -j%NUMBER_OF_PROCESSORS%
    jom -j%NUMBER_OF_PROCESSORS% install
mac:
    find ../../patches/qtbase_$QT -type f -print0 | sort -z | xargs -0 git apply
    cd ..

    CONFIGURATIONS=-debug
release:
    CONFIGURATIONS=-debug-and-release
mac:
    ./configure -prefix "$USED_PREFIX/Qt-$QT" \
        $CONFIGURATIONS \
        -force-debug-info \
        -opensource \
        -confirm-license \
        -static \
        -opengl desktop \
        -no-openssl \
        -securetransport \
        -I "$USED_PREFIX/include" \
        LIBJPEG_LIBS="$USED_PREFIX/lib/libjpeg.a" \
        ZLIB_LIBS="$USED_PREFIX/lib/libz.a" \
        -nomake examples \
        -nomake tests \
        -platform macx-clang

    make $MAKE_THREADS_CNT
    make install
""")
else: # qt > '6'
    branch = 'v$QT' + ('-lts-lgpl' if qt < '6.3' else '')
    stage('qt_' + qt, """
    git clone -b """ + branch + """ https://github.com/qt/qt5.git qt_$QT
    cd qt_$QT
    git submodule update --init --recursive qtbase qtimageformats qtsvg
depends:patches/qtbase_""" + qt + """/*.patch
    cd qtbase
mac:
    find ../../patches/qtbase_$QT -type f -print0 | sort -z | xargs -0 git apply -v
    cd ..
    sed -i.bak 's/tqtc-//' {qtimageformats,qtsvg}/dependencies.yaml

    CONFIGURATIONS=-debug
release:
    CONFIGURATIONS=-debug-and-release
mac:
    ./configure -prefix "$USED_PREFIX/Qt-$QT" \
        $CONFIGURATIONS \
        -force-debug-info \
        -opensource \
        -confirm-license \
        -static \
        -opengl desktop \
        -no-openssl \
        -securetransport \
        -system-webp \
        -I "$USED_PREFIX/include" \
        -no-feature-futimens \
        -no-feature-brotli \
        -nomake examples \
        -nomake tests \
        -platform macx-clang -- \
        -DCMAKE_OSX_ARCHITECTURES="x86_64;arm64" \
        -DCMAKE_PREFIX_PATH="$USED_PREFIX"

    ninja
    ninja install
win:
    for /r %%i in (..\\..\\patches\\qtbase_%QT%\\*) do git apply %%i -v
    cd ..

    SET CONFIGURATIONS=-debug
release:
    SET CONFIGURATIONS=-debug-and-release
win:
    """ + removeDir('"%LIBS_DIR%\\Qt' + qt + '"') + """
    SET MOZJPEG_DIR=%LIBS_DIR%\\mozjpeg
    SET OPENSSL_DIR=%LIBS_DIR%\\openssl3
    SET OPENSSL_LIBS_DIR=%OPENSSL_DIR%\\out
    SET ZLIB_LIBS_DIR=%LIBS_DIR%\\zlib
    SET WEBP_DIR=%LIBS_DIR%\\libwebp
    configure -prefix "%LIBS_DIR%\\Qt-%QT%" ^
        %CONFIGURATIONS% ^
        -force-debug-info ^
        -opensource ^
        -confirm-license ^
        -static ^
        -static-runtime ^
        -feature-c++20 ^
        -openssl linked ^
        -system-webp ^
        -system-zlib ^
        -system-libjpeg ^
        -nomake examples ^
        -nomake tests ^
        -platform win32-msvc ^
        -D ZLIB_WINAPI ^
        -- ^
        -D OPENSSL_FOUND=1 ^
        -D OPENSSL_INCLUDE_DIR="%OPENSSL_DIR%\\include" ^
        -D LIB_EAY_DEBUG="%OPENSSL_LIBS_DIR%.dbg\\libcrypto.lib" ^
        -D SSL_EAY_DEBUG="%OPENSSL_LIBS_DIR%.dbg\\libssl.lib" ^
        -D LIB_EAY_RELEASE="%OPENSSL_LIBS_DIR%\\libcrypto.lib" ^
        -D SSL_EAY_RELEASE="%OPENSSL_LIBS_DIR%\\libssl.lib" ^
        -D JPEG_FOUND=1 ^
        -D JPEG_INCLUDE_DIR="%MOZJPEG_DIR%" ^
        -D JPEG_LIBRARY_DEBUG="%MOZJPEG_DIR%\\Debug\\jpeg-static.lib" ^
        -D JPEG_LIBRARY_RELEASE="%MOZJPEG_DIR%\\Release\\jpeg-static.lib" ^
        -D ZLIB_FOUND=1 ^
        -D ZLIB_INCLUDE_DIR="%ZLIB_LIBS_DIR%" ^
        -D ZLIB_LIBRARY_DEBUG="%ZLIB_LIBS_DIR%\\Debug\\zlibstaticd.lib" ^
        -D ZLIB_LIBRARY_RELEASE="%ZLIB_LIBS_DIR%\\Release\\zlibstatic.lib" ^
        -D WebP_INCLUDE_DIR="%WEBP_DIR%\\src" ^
        -D WebP_demux_INCLUDE_DIR="%WEBP_DIR%\\src" ^
        -D WebP_mux_INCLUDE_DIR="%WEBP_DIR%\\src" ^
        -D WebP_LIBRARY="%WEBP_DIR%\\out\\release-static\\$X8664\\lib\\webp.lib" ^
        -D WebP_demux_LIBRARY="%WEBP_DIR%\\out\\release-static\\$X8664\\lib\\webpdemux.lib" ^
        -D WebP_mux_LIBRARY="%WEBP_DIR%\\out\\release-static\\$X8664\\lib\\webpmux.lib"

    cmake --build . --config Debug --parallel
    cmake --install . --config Debug
    cmake --build . --parallel
    cmake --install .
""")

stage('tg_owt', """
    git clone https://github.com/desktop-app/tg_owt.git
    cd tg_owt
    git checkout 4a60ce1ab9
    git submodule init
    git submodule update
win:
    SET MOZJPEG_PATH=$LIBS_DIR/mozjpeg
    SET OPUS_PATH=$USED_PREFIX/include/opus
    SET OPENSSL_PATH=$LIBS_DIR/openssl3/include
    SET LIBVPX_PATH=$USED_PREFIX/include
    SET OPENH264_PATH=$USED_PREFIX/include
    SET FFMPEG_PATH=$LIBS_DIR/ffmpeg
    mkdir out
    cd out
    mkdir Debug
    cd Debug
    cmake -G Ninja \
        -DCMAKE_BUILD_TYPE=Debug \
        -DTG_OWT_BUILD_AUDIO_BACKENDS=OFF \
        -DTG_OWT_SPECIAL_TARGET=$SPECIAL_TARGET \
        -DTG_OWT_LIBJPEG_INCLUDE_PATH=$MOZJPEG_PATH \
        -DTG_OWT_OPENSSL_INCLUDE_PATH=$OPENSSL_PATH \
        -DTG_OWT_OPUS_INCLUDE_PATH=$OPUS_PATH \
        -DTG_OWT_LIBVPX_INCLUDE_PATH=$LIBVPX_PATH \
        -DTG_OWT_OPENH264_INCLUDE_PATH=$OPENH264_PATH \
        -DTG_OWT_FFMPEG_INCLUDE_PATH=$FFMPEG_PATH ../..
    ninja
release:
    cd ..
    mkdir Release
    cd Release
    cmake -G Ninja \
        -DCMAKE_BUILD_TYPE=Release \
        -DTG_OWT_BUILD_AUDIO_BACKENDS=OFF \
        -DTG_OWT_SPECIAL_TARGET=$SPECIAL_TARGET \
        -DTG_OWT_LIBJPEG_INCLUDE_PATH=$MOZJPEG_PATH \
        -DTG_OWT_OPENSSL_INCLUDE_PATH=$OPENSSL_PATH \
        -DTG_OWT_OPUS_INCLUDE_PATH=$OPUS_PATH \
        -DTG_OWT_LIBVPX_INCLUDE_PATH=$LIBVPX_PATH \
        -DTG_OWT_OPENH264_INCLUDE_PATH=$OPENH264_PATH \
        -DTG_OWT_FFMPEG_INCLUDE_PATH=$FFMPEG_PATH ../..
    ninja
mac:
    MOZJPEG_PATH=$USED_PREFIX/include
    OPUS_PATH=$USED_PREFIX/include/opus
    LIBVPX_PATH=$USED_PREFIX/include
    OPENH264_PATH=$USED_PREFIX/include
    FFMPEG_PATH=$USED_PREFIX/include
    mkdir out
    cd out
    mkdir Debug.x86_64
    cd Debug.x86_64
    cmake -G Ninja \
        -DCMAKE_BUILD_TYPE=Debug \
        -DCMAKE_OSX_ARCHITECTURES=x86_64 \
        -DTG_OWT_BUILD_AUDIO_BACKENDS=OFF \
        -DTG_OWT_SPECIAL_TARGET=$SPECIAL_TARGET \
        -DTG_OWT_LIBJPEG_INCLUDE_PATH=$MOZJPEG_PATH \
        -DTG_OWT_OPENSSL_INCLUDE_PATH=$LIBS_DIR/openssl3/include \
        -DTG_OWT_OPUS_INCLUDE_PATH=$OPUS_PATH \
        -DTG_OWT_LIBVPX_INCLUDE_PATH=$LIBVPX_PATH \
        -DTG_OWT_OPENH264_INCLUDE_PATH=$OPENH264_PATH \
        -DTG_OWT_FFMPEG_INCLUDE_PATH=$FFMPEG_PATH ../..
    ninja
    cd ..
    mkdir Debug.arm64
    cd Debug.arm64
    cmake -G Ninja \
        -DCMAKE_BUILD_TYPE=Debug \
        -DCMAKE_OSX_ARCHITECTURES=arm64 \
        -DTG_OWT_BUILD_AUDIO_BACKENDS=OFF \
        -DTG_OWT_SPECIAL_TARGET=$SPECIAL_TARGET \
        -DTG_OWT_LIBJPEG_INCLUDE_PATH=$MOZJPEG_PATH \
        -DTG_OWT_OPENSSL_INCLUDE_PATH=$LIBS_DIR/openssl3/include \
        -DTG_OWT_OPUS_INCLUDE_PATH=$OPUS_PATH \
        -DTG_OWT_LIBVPX_INCLUDE_PATH=$LIBVPX_PATH \
        -DTG_OWT_OPENH264_INCLUDE_PATH=$OPENH264_PATH \
        -DTG_OWT_FFMPEG_INCLUDE_PATH=$FFMPEG_PATH ../..
    ninja
    cd ..
    mkdir Debug
    lipo -create Debug.arm64/libtg_owt.a Debug.x86_64/libtg_owt.a -output Debug/libtg_owt.a
release:
    mkdir Release.x86_64
    cd Release.x86_64
    cmake -G Ninja \
        -DCMAKE_BUILD_TYPE=Release \
        -DCMAKE_OSX_ARCHITECTURES=x86_64 \
        -DTG_OWT_SPECIAL_TARGET=$SPECIAL_TARGET \
        -DTG_OWT_LIBJPEG_INCLUDE_PATH=$MOZJPEG_PATH \
        -DTG_OWT_OPENSSL_INCLUDE_PATH=$LIBS_DIR/openssl3/include \
        -DTG_OWT_OPUS_INCLUDE_PATH=$OPUS_PATH \
        -DTG_OWT_LIBVPX_INCLUDE_PATH=$LIBVPX_PATH \
        -DTG_OWT_OPENH264_INCLUDE_PATH=$OPENH264_PATH \
        -DTG_OWT_FFMPEG_INCLUDE_PATH=$FFMPEG_PATH ../..
    ninja
    cd ..
    mkdir Release.arm64
    cd Release.arm64
    cmake -G Ninja \
        -DCMAKE_BUILD_TYPE=Release \
        -DCMAKE_OSX_ARCHITECTURES=arm64 \
        -DTG_OWT_SPECIAL_TARGET=$SPECIAL_TARGET \
        -DTG_OWT_LIBJPEG_INCLUDE_PATH=$MOZJPEG_PATH \
        -DTG_OWT_OPENSSL_INCLUDE_PATH=$LIBS_DIR/openssl3/include \
        -DTG_OWT_OPUS_INCLUDE_PATH=$OPUS_PATH \
        -DTG_OWT_LIBVPX_INCLUDE_PATH=$LIBVPX_PATH \
        -DTG_OWT_OPENH264_INCLUDE_PATH=$OPENH264_PATH \
        -DTG_OWT_FFMPEG_INCLUDE_PATH=$FFMPEG_PATH ../..
    ninja
    cd ..
    mkdir Release
    lipo -create Release.arm64/libtg_owt.a Release.x86_64/libtg_owt.a -output Release/libtg_owt.a
""")

stage('ada', """
    git clone -b v2.9.0 https://github.com/ada-url/ada.git
    cd ada
win:
    cmake -B out . ^
        -A %WIN32X64% ^
        -D ADA_TESTING=OFF ^
        -D ADA_TOOLS=OFF ^
        -D CMAKE_MSVC_RUNTIME_LIBRARY="MultiThreaded$<$<CONFIG:Debug>:Debug>" ^
        -D CMAKE_C_FLAGS_DEBUG="/MTd /Zi /Ob0 /Od /RTC1" ^
        -D CMAKE_C_FLAGS_RELEASE="/MT /O2 /Ob2 /DNDEBUG"
    cmake --build out --config Debug --parallel
    cmake --build out --config Release --parallel
mac:
    CFLAGS="$UNGUARDED" CPPFLAGS="$UNGUARDED" cmake -B build . \\
        -D ADA_TESTING=OFF \\
        -D ADA_TOOLS=OFF \\
        -D CMAKE_OSX_DEPLOYMENT_TARGET:STRING=$MACOSX_DEPLOYMENT_TARGET \\
        -D CMAKE_OSX_ARCHITECTURES="x86_64;arm64" \\
        -D CMAKE_INSTALL_PREFIX:STRING=$USED_PREFIX
    cmake --build build $MAKE_THREADS_CNT
    cmake --install build
""")

stage('protobuf', """
win:
    git clone --recursive -b v21.9 https://github.com/protocolbuffers/protobuf
    cd protobuf
    git clone https://github.com/abseil/abseil-cpp third_party/abseil-cpp
    cd third_party/abseil-cpp
    git checkout 273292d1cf
    cd ../..
    mkdir build
    cd build
    cmake .. ^
        -A %WIN32X64% ^
        -Dprotobuf_BUILD_TESTS=OFF ^
        -Dprotobuf_BUILD_PROTOBUF_BINARIES=ON ^
        -Dprotobuf_BUILD_LIBPROTOC=ON ^
        -Dprotobuf_WITH_ZLIB_DEFAULT=OFF ^
        -Dprotobuf_DEBUG_POSTFIX=""
    cmake --build . --config Release --parallel
    cmake --build . --config Debug --parallel
""")
# mac:
#     git clone --recursive -b v21.9 https://github.com/protocolbuffers/protobuf
#     cd protobuf
#     git clone https://github.com/abseil/abseil-cpp third_party/abseil-cpp
#     cd third_party/abseil-cpp
#     git checkout 273292d1cf
#     cd ../..
#     mkdir build
#     cd build
#     CFLAGS="$UNGUARDED" CPPFLAGS="$UNGUARDED" cmake .. \
#         -DCMAKE_OSX_DEPLOYMENT_TARGET:STRING=$MACOSX_DEPLOYMENT_TARGET \
#         -Dprotobuf_BUILD_TESTS=OFF \
#         -Dprotobuf_BUILD_PROTOBUF_BINARIES=ON \
#         -Dprotobuf_BUILD_LIBPROTOC=ON \
#         -Dprotobuf_WITH_ZLIB_DEFAULT=OFF
#     cmake --build . $MAKE_THREADS_CNT

if win:
    currentCodePage = subprocess.run('chcp', capture_output=True, shell=True, text=True, env=modifiedEnv).stdout.strip().split()[-1]
    subprocess.run('chcp 65001 > nul', shell=True, env=modifiedEnv)
    runStages()
    subprocess.run('chcp ' + currentCodePage + ' > nul', shell=True, env=modifiedEnv)
else:
    runStages()



'''
This file is part of Telegram Desktop,
the official desktop application for the Telegram messaging service.

For license and copyright information please follow this link:
https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
'''
import sys, os, re, subprocess, io

def finish(code):
    global executePath
    os.chdir(executePath)
    sys.exit(code)

if sys.platform == 'win32' and not 'COMSPEC' in os.environ:
    print('[ERROR] COMSPEC environment variable is not set.')
    finish(1)

executePath = os.getcwd()
scriptPath = os.path.dirname(os.path.realpath(__file__))

inputVersion = ''
versionOriginal = ''
versionMajor = ''
versionMinor = ''
versionPatch = ''
versionAlpha = '0'
versionBeta = False
for arg in sys.argv:
  match = re.match(r'^\s*(\d+)\.(\d+)(\.(\d+)(\.(\d+|beta))?)?\s*$', arg)
  if match:
    inputVersion = arg
    versionOriginal = inputVersion
    versionMajor = match.group(1)
    versionMinor = match.group(2)
    versionPatch = match.group(4) if match.group(4) else '0'
    versionAlphaBeta = match.group(5) if match.group(5) else ''
    if len(versionAlphaBeta) > 0:
      if match.group(6) == 'beta':
        versionBeta = True
      else:
        versionAlpha = match.group(6)

if not len(versionMajor):
  print("Wrong version parameter")
  finish(1)

def checkVersionPart(part):
  cleared = int(part) % 1000 if len(part) > 0 else 0
  if str(cleared) != part:
    print("Bad version part: " + part)
    finish(1)

checkVersionPart(versionMajor)
checkVersionPart(versionMinor)
checkVersionPart(versionPatch)
checkVersionPart(versionAlpha)

versionFull = str(int(versionMajor) * 1000000 + int(versionMinor) * 1000 + int(versionPatch))
versionFullAlpha = '0'
if versionAlpha != '0':
  versionFullAlpha = str(int(versionFull) * 1000 + int(versionAlpha))

versionStr = versionMajor + '.' + versionMinor + '.' + versionPatch
versionStrSmall = versionStr if versionPatch != '0' else versionMajor + '.' + versionMinor

if versionBeta:
  print('Setting version: ' + versionStr + ' beta')
elif versionAlpha != '0':
  print('Setting version: ' + versionStr + '.' + versionAlpha + ' closed alpha')
else:
  print('Setting version: ' + versionStr + ' stable')

#def replaceInFile(path, replaces):

def checkChangelog():
  global scriptPath, versionStr, versionStrSmall

  count = 0
  with io.open(scriptPath + '/../../changelog.txt', encoding='utf-8') as f:
    for line in f:
      if line.startswith(versionStr + ' ') or line.startswith(versionStrSmall + ' '):
        count = count + 1
  if count == 0:
    print('Changelog entry not found!')
    finish(1)
  elif count != 1:
    print('Wrong changelog entries count found: ' + count)
    finish(1)

checkChangelog()

def replaceInFile(path, replacements):
  content = ''
  foundReplacements = {}
  updated = False
  with open(path, 'r') as f:
    for line in f:
      for replacement in replacements:
        if re.search(replacement[0], line):
          changed = re.sub(replacement[0], replacement[1], line)
          if changed != line:
            line = changed
            updated = True
          foundReplacements[replacement[0]] = True
      content = content + line
  for replacement in replacements:
    if not replacement[0] in foundReplacements:
      print('Could not find "' + replacement[0] + '" in "' + path + '".')
      finish(1)
  if updated:
    with open(path, 'w') as f:
      f.write(content)

print('Patching build/version...')
replaceInFile(scriptPath + '/version', [
  [ r'(AppVersion\s+)\d+', r'\g<1>' + versionFull ],
  [ r'(AppVersionStrMajor\s+)\d[\d\.]*', r'\g<1>' + versionMajor + '.' + versionMinor ],
  [ r'(AppVersionStrSmall\s+)\d[\d\.]*', r'\g<1>' + versionStrSmall ],
  [ r'(AppVersionStr\s+)\d[\d\.]*', r'\g<1>' + versionStr ],
  [ r'(BetaChannel\s+)\d', r'\g<1>' + ('1' if versionBeta else '0') ],
  [ r'(AlphaVersion\s+)\d+', r'\g<1>' + versionFullAlpha ],
  [ r'(AppVersionOriginal\s+)\d[\d\.beta]*', r'\g<1>' + versionOriginal ],
])

print('Patching core/version.h...')
replaceInFile(scriptPath + '/../SourceFiles/core/version.h', [
  [ r'(TDESKTOP_REQUESTED_ALPHA_VERSION\s+)\(\d+ULL\)', r'\g<1>(' + versionFullAlpha + 'ULL)' ],
  [ r'(AppVersion\s+=\s+)\d+', r'\g<1>' + versionFull ],
  [ r'(AppVersionStr\s+=\s+)[^;]+', r'\g<1>"' + versionStrSmall + '"' ],
  [ r'(AppBetaVersion\s+=\s+)[a-z]+', r'\g<1>' + ('true' if versionBeta else 'false') ],
])

parts = [versionMajor, versionMinor, versionPatch, versionAlpha]
withcomma = ','.join(parts)
withdot = '.'.join(parts)
rcReplaces = [
  [ r'(FILEVERSION\s+)\d+,\d+,\d+,\d+', r'\g<1>' + withcomma ],
  [ r'(PRODUCTVERSION\s+)\d+,\d+,\d+,\d+', r'\g<1>' + withcomma ],
  [ r'("FileVersion",\s+)"\d+\.\d+\.\d+\.\d+"', r'\g<1>"' + withdot + '"' ],
  [ r'("ProductVersion",\s+)"\d+\.\d+\.\d+\.\d+"', r'\g<1>"' + withdot + '"' ],
]

print('Patching Telegram.rc...')
replaceInFile(scriptPath + '/../Resources/winrc/Telegram.rc', rcReplaces)

print('Patching Updater.rc...')
replaceInFile(scriptPath + '/../Resources/winrc/Updater.rc', rcReplaces)

print('Patching appxmanifest.xml...')
replaceInFile(scriptPath + '/../Resources/uwp/AppX/AppxManifest.xml', [
  [ r'( Version=)"\d+\.\d+\.\d+\.\d+"', r'\g<1>"' + withdot + '"' ],
])



import os, sys, re, subprocess, datetime, time

executePath = os.getcwd()
scriptPath = os.path.dirname(os.path.realpath(__file__))

lastCommit = ''
today = ''
uuid = ''
nextLast = False
nextDate = False
nextUuid = False
building = True
composing = False
conf = 'Release'
for arg in sys.argv:
    if nextLast:
        lastCommit = arg
        nextLast = False
    elif nextDate:
        today = arg
        nextDate = False
    elif nextUuid:
        uuid = arg
        nextUuid = False
    elif arg == 'send':
        building = False
        composing = False
    elif arg == 'from':
        nextLast = True
        building = False
        composing = True
    elif arg == 'date':
        nextDate = True
    elif arg == 'request_uuid':
        nextUuid = True
    elif arg == 'debug':
        conf = 'Debug'

def finish(code, error = ''):
    if error != '':
        print('[ERROR] ' + error)
    global executePath
    os.chdir(executePath)
    sys.exit(code)

os.chdir(scriptPath + '/..')

if 'AC_USERNAME' not in os.environ:
    finish(1, 'AC_USERNAME not found!')
username = os.environ['AC_USERNAME']

if today == '':
    today = datetime.datetime.now().strftime("%d_%m_%y")
outputFolder = 'updates/' + today

archive = 'tdesktop_macOS_' + today + '.zip'

if building:
    print('Building ' + conf + ' version for OS X 10.13+..')

    if os.path.exists('../out/' + conf + '/' + outputFolder):
        finish(1, 'Todays updates version exists.')

    if uuid == '':
        result = subprocess.call('./configure.sh', shell=True)
        if result != 0:
            finish(1, 'While calling GYP.')

    os.chdir('../out')
    if uuid == '':
        result = subprocess.call('cmake --build . --config ' + conf + ' --target Telegram', shell=True)
        if result != 0:
            finish(1, 'While building Telegram.')

    os.chdir(conf);
    if uuid == '':
        if not os.path.exists('Telegram.app'):
            finish(1, 'Telegram.app not found.')

        result = subprocess.call('strip Telegram.app/Contents/MacOS/Telegram', shell=True)
        if result != 0:
            finish(1, 'While stripping Telegram.')

        result = subprocess.call('codesign --force --deep --timestamp --options runtime --sign "Developer ID Application: Telegram FZ-LLC (C67CF9S4VU)" Telegram.app --entitlements "../../Telegram/Telegram/Telegram.entitlements"', shell=True)
        if result != 0:
            finish(1, 'While signing Telegram.')

        if not os.path.exists('Telegram.app/Contents/Frameworks/Updater'):
            finish(1, 'Updater not found.')
        elif not os.path.exists('Telegram.app/Contents/Helpers/crashpad_handler'):
            finish(1, 'crashpad_handler not found.')
        elif not os.path.exists('Telegram.app/Contents/Resources/Icon.icns'):
            finish(1, 'Icon not found.')
        elif not os.path.exists('Telegram.app/Contents/_CodeSignature'):
            finish(1, 'Signature not found.')

        if os.path.exists(today):
            subprocess.call('rm -rf ' + today, shell=True)
        result = subprocess.call('mkdir -p ' + today + '/TelegramForcePortable', shell=True)
        if result != 0:
            finish(1, 'Creating folder ' + today + '/TelegramForcePortable')

        result = subprocess.call('cp -r Telegram.app ' + today + '/', shell=True)
        if result != 0:
            finish(1, 'Cloning Telegram.app to ' + today + '.')

        result = subprocess.call('zip -r ' + archive + ' ' + today, shell=True)
        if result != 0:
            finish(1, 'Adding tdesktop to archive.')

        print('Beginning notarization process.')
        result = subprocess.call('xcrun notarytool submit "' + archive + '" --keychain-profile "preston" --wait', shell=True)
        if result != 0:
            finish(1, 'Notarizing the archive.')
    result = subprocess.call('xcrun stapler staple Telegram.app', shell=True)
    if result != 0:
        finish(1, 'Error calling stapler')

    subprocess.call('rm -rf ' + today + '/Telegram.app', shell=True)
    subprocess.call('rm ' + archive, shell=True)
    result = subprocess.call('cp -r Telegram.app ' + today + '/', shell=True)
    if result != 0:
        finish(1, 'Re-Cloning Telegram.app to ' + today + '.')

    result = subprocess.call('zip -r ' + archive + ' ' + today, shell=True)
    if result != 0:
        finish(1, 'Re-Adding tdesktop to archive.')
    print('Re-Archived.')

    subprocess.call('mkdir -p ' + outputFolder, shell=True)
    subprocess.call('mv ' + archive + ' ' + outputFolder + '/', shell=True)
    subprocess.call('rm -rf ' + today, shell=True)
    print('Finished.')
    finish(0)

commandPath = scriptPath + '/../../out/' + conf + '/' + outputFolder + '/command.txt'

if composing:
    templatePath = scriptPath + '/../../../DesktopPrivate/updates_template.txt'
    if not os.path.exists(templatePath):
        finish(1, 'Template file "' + templatePath + '" not found.')

    if not re.match(r'^[a-f0-9]{9,40}$', lastCommit):
        finish(1, 'Wrong last commit: ' + lastCommit)

    log = subprocess.check_output(['git', 'log', lastCommit+'..HEAD']).decode('utf-8')
    logLines = log.split('\n')
    firstCommit = ''
    commits = []
    for line in logLines:
        if line.startswith('commit '):
            commit = line.split(' ')[1]
            if not len(firstCommit):
                firstCommit = commit
            commits.append('')
        elif line.startswith('    '):
            stripped = line[4:]
            if not len(stripped):
                continue
            elif not len(commits):
                print(log)
                finish(1, 'Bad git log output.')
            if len(commits[len(commits) - 1]):
                commits[len(commits) - 1] += '\n' + stripped
            else:
                commits[len(commits) - 1] = '- ' + stripped
    commits.reverse()
    if not len(commits):
        finish(1, 'No commits since last build :(')

    changelog = '\n'.join(commits)
    print('\n\nReady! File: ' + archive + '\nChangelog:\n' + changelog)
    with open(templatePath, 'r') as template:
        with open(commandPath, 'w') as f:
            for line in template:
                if line.startswith('//'):
                    continue
                line = line.replace('{path}', scriptPath + '/../../out/' + conf + '/' + outputFolder + '/' + archive)
                line = line.replace('{caption}', 'TDesktop at ' + today.replace('_', '.') + ':\n\n' + changelog)
                f.write(line)
    print('\n\nEdit:\n')
    print('vi ' + commandPath)
    finish(0)

if not os.path.exists(commandPath):
    finish(1, 'Command file not found.')

readingCaption = False
caption = ''
with open(commandPath, 'r') as f:
    for line in f:
        if readingCaption:
            caption = caption + line
        elif line.startswith('caption: '):
            readingCaption = True
            caption = line[len('caption: '):]
            if not caption.startswith('TDesktop at ' + today.replace('_', '.') + ':'):
                finish(1, 'Wrong caption start.')
print('\n\nSending! File: ' + archive + '\nChangelog:\n' + caption)
if len(caption) > 1024:
    print('Length: ' + str(len(caption)))
    print('vi ' + commandPath)
    finish(1, 'Too large.')

if not os.path.exists('../out/' + conf + '/' + outputFolder + '/' + archive):
    finish(1, 'Not built yet.')

subprocess.call(scriptPath + '/../../out/' + conf + '/Telegram.app/Contents/MacOS/Telegram -sendpath interpret://' + scriptPath + '/../../out/' + conf + '/' + outputFolder + '/command.txt', shell=True)

finish(0)



'''
This file is part of Telegram Desktop,
the official desktop application for the Telegram messaging service.

For license and copyright information please follow this link:
https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
'''
import glob, re, binascii, os, sys

sys.dont_write_bytecode = True
scriptPath = os.path.dirname(os.path.realpath(__file__))
sys.path.append(scriptPath + '/../../../lib_tl/tl')
from generate_tl import generate

generate({
  'namespaces': {
    'creator': 'MTP::details',
  },
  'prefixes': {
    'type': 'MTP',
    'data': 'MTPD',
    'id': 'mtpc',
    'construct': 'MTP_',
  },
  'types': {
    'prime': 'mtpPrime',
    'typeId': 'mtpTypeId',
    'buffer': 'mtpBuffer',
  },
  'sections': [
    'read-write',
  ],

  # define some checked flag conversions
  # the key flag type should be a subset of the value flag type
  # with exact the same names, then the key flag can be implicitly
  # casted to the value flag type
  'flagInheritance': {
    'messageService': 'message',
    'updateShortMessage': 'message',
    'updateShortChatMessage': 'message',
    'updateShortSentMessage': 'message',
    'replyKeyboardHide': 'replyKeyboardMarkup',
    'replyKeyboardForceReply': 'replyKeyboardMarkup',
    'inputPeerNotifySettings': 'peerNotifySettings',
    'peerNotifySettings': 'inputPeerNotifySettings',
    'channelForbidden': 'channel',
    'dialogFolder': 'dialog',
  },

  'typeIdExceptions': [
    'channel#c88974ac',
    'ipPortSecret#37982646',
    'accessPointRule#4679b65f',
    'help.configSimple#5a592a6c',
    'messageReplies#81834865',
  ],

  'renamedTypes': {
    'passwordKdfAlgoSHA256SHA256PBKDF2HMACSHA512iter100000SHA256ModPow': 'passwordKdfAlgoModPow',
   },

  'skip': [
    'int ? = Int;',
    'long ? = Long;',
    'double ? = Double;',
    'string ? = String;',

    'vector {t:Type} # [ t ] = Vector t;',

    'int128 4*[ int ] = Int128;',
    'int256 8*[ int ] = Int256;',

    'vector#1cb5c415 {t:Type} # [ t ] = Vector t;',
  ],
  'builtin': [
    'int',
    'long',
    'double',
    'string',
    'bytes',
    'int128',
    'int256',
  ],
  'builtinTemplates': [
    'vector',
    'flags',
  ],
  'synonyms': {
    'bytes': 'string',
  },
  'builtinInclude': 'mtproto/core_types.h',
  'optimizeSingleData': True,

  'dumpToText': {
    'include': 'mtproto/details/mtproto_dump_to_text.h',
  },

})



# This file is part of Desktop App Toolkit,
# a set of libraries for developing nice desktop applications.
#
# For license and copyright information please follow this link:
# https://github.com/desktop-app/legal/blob/master/LEGAL

from __future__ import print_function
import sys, os, re, hashlib

def error(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
    sys.exit(1)

try:
    from win32api import GetFileVersionInfo
except ImportError:
    error('Python module "pywin32" is not installed.\n\nTry "' + sys.executable + ' -m pip install pywin32".')

if len(sys.argv) < 2:
    error('Expected input path to "d3dcompiler_47.dll".')

inputPath = sys.argv[1]
if not os.path.exists(inputPath):
    error('File "' + inputPath + '" doesn\'t exist.')

info = GetFileVersionInfo(inputPath, '\\')
version = [ info['FileVersionMS'] // 65536, info['FileVersionMS'] % 65536, info['FileVersionLS'] // 65536, info['FileVersionLS'] % 65536 ]
if (version != [10, 0, 22621, 3233]):
    error('Bad "d3dcompiler_47.dll" version: ' + '.'.join(str(x) for x in version))

bufferSize = 1024 * 1024
sha256 = hashlib.sha256()

with open(inputPath, 'rb') as f:
    print(hashlib.sha256(f.read()).hexdigest())



import os, sys, requests, re
from os.path import expanduser

filename = ''
for arg in sys.argv:
  if re.match(r'.*\.strings$', arg):
    filename = expanduser(arg)

result = ''
if not os.path.isfile(filename):
  print("File not found.")
  sys.exit(1)

with open(filename) as f:
  for line in f:
    if re.match(r'\s*\/\* new from [a-zA-Z0-9\.]+ \*\/\s*', line):
      continue
    if re.match(r'\"lng_[a-z_]+\#(zero|two|few|many)\".+', line):
      continue
    result = result + line

remove = 0
while (len(result) > remove + 1) and (result[len(result) - remove - 1] == '\n') and (result[len(result) - remove - 2] == '\n'):
  remove = remove + 1
result = result[:len(result) - remove]

with open('lang.strings', 'w') as out:
  out.write(result)

sys.exit()



import os, sys, requests, pprint, re, json
from uritemplate import URITemplate, expand
from subprocess import call, Popen, PIPE
from os.path import expanduser

changelog_file = '../../changelog.txt'
token_file = '../../../DesktopPrivate/github-releases-token.txt'

version = ''
commit = ''
for arg in sys.argv:
  if re.match(r'\d+\.\d+', arg):
    version = arg
  elif re.match(r'^[a-f0-9]{40}$', arg):
    commit = arg

# thanks http://stackoverflow.com/questions/13909900/progress-of-python-requests-post
class upload_in_chunks(object):
  def __init__(self, filename, chunksize=1 << 13):
    self.filename = filename
    self.chunksize = chunksize
    self.totalsize = os.path.getsize(filename)
    self.readsofar = 0

  def __iter__(self):
    with open(self.filename, 'rb') as file:
      while True:
        data = file.read(self.chunksize)
        if not data:
          sys.stderr.write("\n")
          break
        self.readsofar += len(data)
        percent = self.readsofar * 1e2 / self.totalsize
        sys.stderr.write("\r{percent:3.0f}%".format(percent=percent))
        yield data

  def __len__(self):
    return self.totalsize

class IterableToFileAdapter(object):
  def __init__(self, iterable):
    self.iterator = iter(iterable)
    self.length = len(iterable)

  def read(self, size=-1): # TBD: add buffer for `len(data) > size` case
    return next(self.iterator, b'')

  def __len__(self):
    return self.length

def checkResponseCode(result, right_code):
  if (result.status_code != right_code):
    print('Wrong result code: ' + str(result.status_code) + ', should be ' + str(right_code))
    sys.exit(1)

def getOutput(command):
  p = Popen(command.split(), stdout=PIPE)
  output, err = p.communicate()
  if err != None or p.returncode != 0:
    print('ERROR!')
    print(err)
    print(p.returncode)
    sys.exit(1)
  return output.decode('utf-8')

def invoke(command):
  return call(command.split()) == 0

def appendSubmodules(appendTo, root, rootRevision):
  startpath = os.getcwd()
  lines = getOutput('git submodule foreach').split('\n')
  for line in lines:
    if len(line) == 0:
      continue
    match = re.match(r"^Entering '([^']+)'$", line)
    if not match:
      print('Bad line: ' + line)
      return False
    path = match.group(1)
    subroot = root + '/' + path
    revision = getOutput('git rev-parse ' + rootRevision + ':' + path).split('\n')[0]
    print('Adding submodule ' + path + '...')
    os.chdir(path)
    tmppath = appendTo + '_tmp'
    if not invoke('git archive --prefix=' + subroot + '/ ' + revision + ' -o ' + tmppath + '.tar'):
      os.remove(appendTo + '.tar')
      os.remove(tmppath + '.tar')
      return False
    if not appendSubmodules(tmppath, subroot, revision):
      return False
    tar = 'tar' if sys.platform == 'linux' else 'gtar'
    if not invoke(tar + ' --concatenate --file=' + appendTo + '.tar ' + tmppath + '.tar'):
      os.remove(appendTo + '.tar')
      os.remove(tmppath + '.tar')
      return False
    os.remove(tmppath + '.tar')
    os.chdir(startpath)
  return True

def prepareSources():
  workpath = os.getcwd()
  os.chdir('../..')
  rootpath = os.getcwd()
  finalpart = rootpath + '/out/Release/sources'
  finalpath = finalpart + '.tar'
  if os.path.exists(finalpath):
    os.remove(finalpath)
  if os.path.exists(finalpath + '.gz'):
    os.remove(finalpath + '.gz')
  tmppath = rootpath + '/out/Release/tmp.tar'
  print('Preparing source tarball...')
  revision = 'v' + version
  targetRoot = 'tdesktop-' + version + '-full';
  if not invoke('git archive --prefix=' + targetRoot + '/ -o ' + finalpath + ' ' + revision):
    os.remove(finalpath)
    sys.exit(1)
  if not appendSubmodules(finalpart, targetRoot, revision):
    sys.exit(1)
  print('Compressing...')
  if not invoke('gzip -9 ' + finalpath):
    os.remove(finalpath)
    sys.exit(1)
  os.chdir(workpath)
  return finalpath + '.gz'

pp = pprint.PrettyPrinter(indent=2)
url = 'https://api.github.com/'

version_parts = version.split('.')

stable = 1
beta = 0

if len(version_parts) < 2:
  print('Error: expected at least major version ' + version)
  sys.exit(1)
if len(version_parts) > 4:
  print('Error: bad version passed ' + version)
  sys.exit(1)
version_major = version_parts[0] + '.' + version_parts[1]
if len(version_parts) == 2:
  version = version_major + '.0'
  version_full = version
else:
  version = version_major + '.' + version_parts[2]
  version_full = version
  if len(version_parts) == 4:
    if version_parts[3] == 'beta':
      beta = 1
      stable = 0
      version_full = version + '.beta'
    else:
      print('Error: unexpected version part ' + version_parts[3])
      sys.exit(1)

access_token = ''
if os.path.isfile(token_file):
  with open(token_file) as f:
    for line in f:
      access_token = line.replace('\n', '')

if access_token == '':
  print('Access token not found!')
  sys.exit(1)

print('Version: ' + version_full)
local_base = expanduser("~") + '/Projects/backup/tdesktop'
if not os.path.isdir(local_base):
    local_base = '/mnt/c/Telegram/Projects/backup/tdesktop'
    if not os.path.isdir(local_base):
        print('Backup path not found: ' + local_base)
        sys.exit(1)
local_folder = local_base + '/' + version_major + '/' + version_full

if stable == 1:
  if os.path.isdir(local_folder + '.beta'):
    beta = 1
    stable = 0
    version_full = version + '.beta'
    local_folder = local_folder + '.beta'

if not os.path.isdir(local_folder):
  print('Storage path not found: ' + local_folder)
  sys.exit(1)

local_folder = local_folder + '/'

files = []
files.append({
  'local': 'sources',
  'remote': 'tdesktop-' + version + '-full.tar.gz',
  'mime': 'application/x-gzip',
  'label': 'Source code (tar.gz, full)',
})
files.append({
  'local': 'tsetup.' + version_full + '.exe',
  'remote': 'tsetup.' + version_full + '.exe',
  'backup_folder': 'tsetup',
  'mime': 'application/octet-stream',
  'label': 'Windows 32 bit: Installer',
})
files.append({
  'local': 'tportable.' + version_full + '.zip',
  'remote': 'tportable.' + version_full + '.zip',
  'backup_folder': 'tsetup',
  'mime': 'application/zip',
  'label': 'Windows 32 bit: Portable',
})
files.append({
  'local': 'tsetup-x64.' + version_full + '.exe',
  'remote': 'tsetup-x64.' + version_full + '.exe',
  'backup_folder': 'tx64',
  'mime': 'application/octet-stream',
  'label': 'Windows 64 bit: Installer',
})
files.append({
  'local': 'tportable-x64.' + version_full + '.zip',
  'remote': 'tportable-x64.' + version_full + '.zip',
  'backup_folder': 'tx64',
  'mime': 'application/zip',
  'label': 'Windows 64 bit: Portable',
})
files.append({
  'local': 'tsetup-arm64.' + version_full + '.exe',
  'remote': 'tsetup-arm64.' + version_full + '.exe',
  'backup_folder': 'tarm64',
  'mime': 'application/octet-stream',
  'label': 'Windows on ARM: Installer',
})
files.append({
  'local': 'tportable-arm64.' + version_full + '.zip',
  'remote': 'tportable-arm64.' + version_full + '.zip',
  'backup_folder': 'tarm64',
  'mime': 'application/zip',
  'label': 'Windows on ARM: Portable',
})
files.append({
  'local': 'tsetup.' + version_full + '.dmg',
  'remote': 'tsetup.' + version_full + '.dmg',
  'backup_folder': 'tmac',
  'mime': 'application/octet-stream',
  'label': 'macOS 10.13+: Installer',
})
files.append({
  'local': 'tsetup.' + version_full + '.tar.xz',
  'remote': 'tsetup.' + version_full + '.tar.xz',
  'backup_folder': 'tlinux',
  'mime': 'application/octet-stream',
  'label': 'Linux 64 bit: Binary',
})

r = requests.get(url + 'repos/telegramdesktop/tdesktop/releases/tags/v' + version)
if r.status_code == 404:
  print('Release not found, creating.')
  if commit == '':
    print('Error: specify the commit.')
    sys.exit(1)
  if not os.path.isfile(changelog_file):
    print('Error: Changelog file not found.')
    sys.exit(1)
  changelog = ''
  started = 0
  with open(changelog_file) as f:
    for line in f:
      if started == 1:
        if re.match(r'^\d+\.\d+', line):
          break
        changelog += line
      else:
        if re.match(r'^\d+\.\d+', line):
          if line[0:len(version) + 1] == version + ' ':
            started = 1
          elif line[0:len(version_major) + 1] == version_major + ' ':
            if version_major + '.0' == version:
              started = 1
  if started != 1:
    print('Error: Changelog not found.')
    sys.exit(1)

  changelog = changelog.strip()
  print('Changelog: ')
  print(changelog)

  r = requests.post(url + 'repos/telegramdesktop/tdesktop/releases', headers={'Authorization': 'token ' + access_token}, data=json.dumps({
    'tag_name': 'v' + version,
    'target_commitish': commit,
    'name': 'v ' + version,
    'body': changelog,
    'prerelease': (beta == 1),
  }))
  checkResponseCode(r, 201)

tagname = 'v' + version
invoke("git fetch origin")
if stable == 1:
  invoke("git push launchpad {}:master".format(tagname))
else:
  invoke("git push launchpad {}:beta".format(tagname))
invoke("git push --tags launchpad")

r = requests.get(url + 'repos/telegramdesktop/tdesktop/releases/tags/v' + version)
checkResponseCode(r, 200)

release_data = r.json()
#pp.pprint(release_data)

release_id = release_data['id']
print('Release ID: ' + str(release_id))

r = requests.get(url + 'repos/telegramdesktop/tdesktop/releases/' + str(release_id) + '/assets')
checkResponseCode(r, 200)

assets = release_data['assets']
for asset in assets:
  name = asset['name']
  found = 0
  for file in files:
    if file['remote'] == name:
      print('Already uploaded: ' + name)
      file['already'] = 1
      found = 1
      break
  if found == 0:
    print('Warning: strange asset: ' + name)

for file in files:
  if 'already' in file:
    continue
  if file['local'] == 'sources':
    file_path = prepareSources()
  else:
    file_path = local_folder + file['backup_folder'] + '/' + file['local']
  if not os.path.isfile(file_path):
    print('Warning: file not found ' + file['local'])
    continue

  upload_url = expand(release_data['upload_url'], {'name': file['remote'], 'label': file['label']})

  content = upload_in_chunks(file_path, 10)

  print('Uploading: ' + file['remote'] + ' (' + str(round(len(content) / 10000) / 100.) + ' MB)')
  r = requests.post(upload_url, headers={'Content-Type': file['mime'], 'Authorization': 'token ' + access_token}, data=IterableToFileAdapter(content))

  checkResponseCode(r, 201)

  print('Success! Removing.')
  if not invoke('rm ' + file_path):
    print('Bad rm return code :(')
    sys.exit(1)

sys.exit()



#!/usr/bin/env python3
from os import environ
from os.path import dirname
from jinja2 import Environment, FileSystemLoader

def checkEnv(envName, defaultValue):
    return bool(len(environ[envName])) if envName in environ else defaultValue

def main():
    print(Environment(loader=FileSystemLoader(dirname(__file__))).get_template("Dockerfile").render(
        DEBUG=checkEnv("DEBUG", True),
        LTO=checkEnv("LTO", True),
    ))

if __name__ == '__main__':
    main()
